"cosmo.fit" <-
  function(seqs, minW, maxW=NULL, revComp=TRUE, minSites=NULL, maxSites = NULL, oops=FALSE, zoops=TRUE, tcm=FALSE, approx="over", cutFac=2, wCrit="eval", wFold=5, wTrunc=100, modCrit="eval", modFold=5, modTrunc=100, conCrit="pwmCV", conFold=2, conTrunc=100, intCrit="eval", intFold=5, intTrunc=100, constraints=NULL, starts=5, lstarts=FALSE, backSeqs=NULL, backFold=5,transMat=NULL, order=NULL){
    
    ## initialize variables
    numSeqs          = ncol(seqs)
    maxSeqLength     = nrow(seqs)
    maxW             = ifelse(is.null(maxW),minW,maxW)
    numBackSeqs      = ifelse(is.null(backSeqs),0,ncol(backSeqs))
    maxBackSeqLength = ifelse(is.null(backSeqs),0,nrow(backSeqs))
    backFold         = ifelse(!is.null(backSeqs) && backFold>numBackSeqs, numBackSeqs, backFold)

  
    ## check arguments:
    if (!is.matrix(seqs) || !is.numeric(seqs)) stop("seqs must be a numeric matrix")
    if (!all(seqs %in% 0:4)) stop("All elements of seqs must be in {0,1,2,3,4}")
    if (maxSeqLength < maxW) stop("maxW cannot be greater than the length of the longest sequence")
    
    if (minW < 2 || minW > 30) stop("minW must be at least 2 and no greater than 30")
    if (!is.null(maxW) && (maxW < 2 || maxW > 30)) stop("maxW must be at least 2 and no greater than 30")
    if (!is.null(maxW) && minW > maxW) stop("minW cannot be greater than maxW")

    if (approx!="over" && approx!="cut" && approx!="tcm") stop("approx must be one of: over, cut, tcm")
    if (wCrit!="lik" && wCrit!="aic" && wCrit!="bic" && wCrit!="eval" && wCrit!="likCV"){
      stop("wCrit must be one of: lik, aic, bic, eval, likCV")
    }
    if (modCrit!="lik" && modCrit!="aic" && modCrit!="bic" && modCrit!="eval" && modCrit!="likCV"){
      stop("modCrit must be one of: lik, aic, bic, eval, likCV")
    }
    if (conCrit!="lik" && conCrit!="aic" && conCrit!="bic" && conCrit!="eval" && conCrit!="likCV" && conCrit!="pwmCV"){
      stop("conCrit must be one of: lik, aic, bic, eval, likCV, pwmCV")
    }
    if (intCrit!="lik" && intCrit!="eval" && intCrit!="likCV"){
      stop("intCrit must be one of: lik, eval, likCV")
    }
    
    if (wTrunc < 1   || wTrunc>100)   stop("wTrunc must be between 1 and 100")
    if (modTrunc < 1 || modTrunc>100) stop("modTrunc must be between 1 and 100")
    if (conTrunc < 1 || conTrunc>100) stop("conTrunc must be between 1 and 100")
    if (intTrunc < 1 || intTrunc>100) stop("intTrunc must be between 1 and 100")
    
    if (wFold < 1   || wFold>numSeqs)   stop("wFold must be between 1 and the number of sequences")
    if (modFold < 1 || modFold>numSeqs) stop("modFold must be between 1 and the number of sequences")
    if (conFold < 1 || conFold>numSeqs) stop("conFold must be between 1 and the number of sequences")
    if (intFold < 1 || intFold>numSeqs) stop("intFold must be between 1 and the number of sequences")
    
    if (is.null(backSeqs)){
      if (backFold < 1 || backFold>numSeqs) stop("backFold must be between 1 and the number of sequences")
    }else{
      if (backFold < 1 || backFold>numBackSeqs) stop("backFold must be between 1 and the number of background sequences")
    }
    
    if (!is.null(order) && (order<0 || order>6)) stop("order must be between 0 and 6")
    if (starts<1 || starts>100) stop("starts must be between 1 and 100")
    if (!is.logical(lstarts)) stop("lstarts must be a logical")
    if (!is.logical(oops)) stop("oops must be a logical")
    if (!is.logical(zoops)) stop("zoops must be a logical")
    if (!is.logical(tcm)) stop("tcm must be a logical")
    
    
    if ((!is.null(backSeqs)) && (!is.matrix(backSeqs) || !is.numeric(backSeqs))) stop("backSeqs must be a numeric matrix")
    if ((!is.null(backSeqs)) && (!all(backSeqs %in% 0:4))) stop("All elements of backSeqs must be in {0,1,2,3,4}")
    

    ## extract transition matrices out of list and concatenate
    ## them in one big matrix, then take the transpose
    ## for compatibility with C
    tM = NULL
    if (!is.null(transMat)){
      for (i in 1:length(transMat)) tM = rbind(tM, transMat[[i]])
      
      print(tM)
      
      if (!is.matrix(tM) || !is.numeric(tM)) stop("transMat must be a numeric matrix")
      totals = apply(tM,1,sum)
      if (!all(abs(totals-1)<1e-8)) stop("All rows of transMat must add up to 1.0")
      correctRows = FALSE
      numRows = 0
      for (i in 0:6){
        numRows = numRows + 4^i
        if (nrow(tM)==numRows){
          correctRows = TRUE
          order = i
        }
      }
      if (!correctRows) stop("transMat must have number of rows equal to sum_{j=0}^k 4^j for k=0,1,2,..., or 6")
      tM = t(tM)
      print(tM)
    }
    
    ## initialize output variables
    selOrder      = 0
    selWidth      = 0
    selModel      = 0
    selConSet     = 0
    selNumSites   = 0
    numModels     = 0
    eval          = 0
    success       = 1;
  
    ## maximal order must be <= 6
    klDivs        = rep(0,7)
    estTransMat   = rep(0,22000)
    
    ## no more than 50 constraint sets
    numW          = maxW - minW + 1
    numMods       = oops + zoops + tcm
    wCrits        = rep(0,50*numW*numMods)
    modCrits      = rep(0,50*numW*numMods)
    conCrits      = rep(0,50*numW*numMods)  
    
    pwm           = rep(0,4*maxW)
    motProbs      = rep(0,numSeqs*maxSeqLength)
    motRevComp    = rep(0,numSeqs*maxSeqLength)
    alignSeqs     = rep(0,numSeqs*maxSeqLength)
    alignStarts   = rep(0,numSeqs*maxSeqLength)

    ## call C function
    com = .C("ccosmo",
      as.integer(seqs),
      as.integer(numSeqs),
      as.integer(maxSeqLength),
      as.character(constraints),
      as.integer(minW),
      as.integer(maxW),
      as.integer(revComp),
      as.integer(minSites),
      as.integer(maxSites),
      as.integer(oops),
      as.integer(zoops),
      as.integer(tcm),
      as.character(approx),
      as.integer(cutFac),
      as.character(wCrit),
      as.integer(wFold),
      as.integer(wTrunc),
      as.character(modCrit),
      as.integer(modFold),
      as.integer(modTrunc),
      as.character(conCrit),
      as.integer(conFold),
      as.integer(conTrunc),
      as.character(intCrit),
      as.integer(intFold),
      as.integer(intTrunc),
      as.integer(starts),
      as.integer(lstarts),
      as.integer(backSeqs),
      as.integer(numBackSeqs),
      as.integer(maxBackSeqLength),
      as.integer(backFold),
      as.double(tM),
      as.integer(order),
      selOrder=as.integer(selOrder),
      klDivs=as.double(klDivs),
      estTransMat=as.double(estTransMat),
      wCrits=as.double(wCrits),
      modCrits=as.double(modCrits),
      conCrits=as.double(conCrits),    
      selWidth=as.integer(selWidth),
      selModel=as.integer(selModel),
      selConSet=as.integer(selConSet),
      selNumSites=as.integer(selNumSites),
      numModels=as.integer(numModels),
      pwm=as.double(pwm),
      motProbs=as.double(motProbs),
      motRevComp=as.integer(motRevComp),
      alignSeqs=as.integer(alignSeqs),
      alignStarts=as.integer(alignStarts),
      eval = as.double(eval),
      success = as.integer(success),
      package="cosmo"
      )
    
    if (!com$success) return(0)
    
    ## prepare output
    out.klDivs      = com$klDivs
    out.klDivs[out.klDivs==0] = NA
    out.numModels   = com$numModels
    out.wCrits      = com$wCrits[1:out.numModels]
    out.modCrits    = com$modCrits[1:out.numModels]
    out.conCrits    = com$conCrits[1:out.numModels]
    out.wCrits[out.wCrits==0]     = NA
    out.modCrits[out.modCrits==0] = NA
    out.conCrits[out.conCrits==0] = NA
    out.conSetNum   = com$selConSet
    out.pwm         = matrix(com$pwm[1:(4*com$selWidth)],4,com$selWidth)
    out.motProbs    = matrix(com$motProbs,maxSeqLength,numSeqs)
    out.motOnRC     = matrix(as.logical(com$motRevComp),maxSeqLength,numSeqs)
    out.alignSeqs   = com$alignSeqs[1:com$selNumSites]
    out.alignStarts = com$alignStarts[1:com$selNumSites]
    out.cons        = paste(apply(out.pwm,2,getConsensus),collapse="")
    out.ic          = apply(out.pwm,2,getInfCon)
    out.modelType   = c("OOPS","ZOOPS","TCM")[com$selModel]
    
    ## extract estimated transition matrices
    ## and create list from them
    first = 1
    tMs = list()
    transMat = com$estTransMat
    for (suborder in com$selOrder:0){
      tM = matrix(transMat[first:(first+4*4^suborder-1)],4^suborder,4, byrow = TRUE)
      first = first + 4*4^suborder
      tMs = c(tMs, mat=list(tM)) 
    }
    names(tMs) = paste("order",com$selOrder:0, sep="")
  
    ## create alignment of predicted sites
    out.sites       = matrix(0,com$selNumSites,com$selWidth)
    for (i in 1:com$selNumSites){
      out.sites[i,] = seqs[com$alignStarts[i]:(com$alignStarts[i]+com$selWidth-1),com$alignSeqs[i]]
    }
    
    ## return list
    list(width=com$selWidth, model=out.modelType,conset=out.conSetNum,pwm=out.pwm, consensus=out.cons, ic = out.ic, eval = com$eval, numSites=com$selNumSites, alignSeqs=out.alignSeqs, alignStarts=out.alignStarts, probs=out.motProbs, revcomp=out.motOnRC, sites = out.sites, order=com$selOrder, klDivs=out.klDivs, wCrits = out.wCrits, modCrits=out.modCrits, conCrits=out.conCrits,transMats=tMs, seqs=seqs)
    
  }

"draw" <-
  function(probs){
    x<-runif(1)
    index = 4
    for (i in 3:1){
      if (x>sum(probs[1:i])){
        break
    }else{
      index = index - 1
    }
    }
    return(index)
  }


"findex" <-
  function(sequence){
    index = 0
    for (i in 1:length(sequence)){
      index = index + (4^(length(sequence)-i))*sequence[i]
    }
    return(index)
  }



"genSeqs.ZOOPS" <-
  function(n,len,width,order,pi, motif, trans){



    starts<-sample(1:(len-width+1),n,replace=TRUE)
    
    len1<-starts-1
    len2<-len-(starts+width-1)
    
    seqs<-NULL
    motifs<-NULL
    for (i in 1:n){
      if (runif(1)<pi){
        thismotif<-sim.motif(motif)
        
        motifs<-rbind(motifs,thismotif)
        newseq<-c(sim.background(len1[i],order,trans),thismotif,sim.background(len2[i],order,trans))
      }else{
        newseq<-sim.background(len,order,trans);
        starts[i]=-1;
      }
      
      seqs<-cbind(seqs,newseq)
    }
    
    list(starts=starts,seqs=seqs)
  }

"getColor" <-
  function(x){
    colors<-x
    colors[colors==1]<-"green"
    colors[colors==2]<-"blue"
    colors[colors==3]<-"orange"
    colors[colors==4]<-"red"
    return(colors)
  }

"getConsensus" <-
  function(col){
    letters = c("A","C","G","T")
    letters[rev(order(col))[1]]
  }

"getInfCon" <-
  function(col){
    ic = 2
    for (p in col){
      if (p!=0) ic = ic + log(p)/log(2)*p
    }
    ic
  }

"getrow" <-
  function(sequence,order){
    if (length(sequence)==0){
      row = 0
      for (i in 1:order){
        row = row + 4^i
      }
    }else{
      fixedrows = 0
      if (length(sequence)<order){
        for (i in (length(sequence)+1):order){
          fixedrows = fixedrows + 4^i
        }
    }
      row = fixedrows+findex(sequence)
    }
    row + 1
  }


"pal" <-
  function(x){
    2 + 2*x*log(x)/log(2) + (1-2*x)*log((1-2*x)/2)/log(2)
  }

"score" <-
  function(comode.out, starts){
    trueStarts = starts[comode.out$alignSeqs]
    correct = comode.out$alignStarts==trueStarts
    
    sens = sum(comode.out$alignStarts==trueStarts)/length(starts)
    spec = sum(comode.out$alignStarts==trueStarts)/length(comode.out$alignStarts) 
    
    list(sens=sens,spec=spec)
  }

"sim.background" <-
  function(len,order,trans){
    if (len==0) return(NULL)
    oligo = NULL
    seq = NULL
    for (i in 1:len){
      probs = as.numeric(trans[getrow(oligo,order),])
      index = draw(probs)
      seq = c(seq,index)
      oligo = c(oligo,index)
      if (length(oligo) > order){
        oligo = oligo[-1]
      }
    }
    seq
  }

"sim.motif" <-
  function(motif){
    seq<-NULL
    for (i in 1:dim(motif)[2]){
      index = draw(as.numeric(motif[,i]))
      seq<-c(seq,index)
    }
    seq
  }


"ic" <-
  function(probs){
    res = log(length(probs))/log(2);
    for (i in 1:length(probs)){
      if (probs[i]>0){
        res = res + (probs[i]*log(probs[i]))/log(2)
      }
    }
    res
  }


".onLoad" <-
  function(lib, pkg){


    require(methods)
  }

            
