/*****************************************************************************
*			     					             *
*      cosmo - Constrained search for motifs in DNA sequences                *
*      Copyright (C) 2006  Oliver Bembom (bembom@berkeley.edu)               *
*                                                                            *
*      This file is part of cosmo.                                           *
*                                                                            *
*      cosmo is free software; you can redistribute it and/or                *
*      modify it under the terms of the GNU Lesser General Public            *
*      License as published by the Free Software Foundation; either          *
*      version 2.1 of the License, or (at your option) any later version.    *
*                                                                            *
*      cosmo is distributed in the hope that it will be useful,              *
*      but WITHOUT ANY WARRANTY; without even the implied warranty of        *
*      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU     *
*      Lesser General Public License for more details.                       *
*                                                                            *
*      You should have received a copy of the GNU Lesser General Public      *
*      License along with this program; if not, write to the Free Software   *
*      Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,            *
*      MA  02110-1301, USA.                                                  *
*                                                                            *
*                                                                            *
*****************************************************************************/


/*****************************************************************************
*								       	     *
*      This module contains the function to be called by R.                  *
*                                                                            *
*****************************************************************************/

#include <R_ext/PrtUtil.h>
#include <R_ext/Memory.h>
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <float.h>
#include <string.h>
#include "general.h"
#include "macros.h"
#include "rcosmo.h"
#include "cosmo.h"
#include "alloc.h"
#include "back.h"
#include "constraints.h"
#include "estarts.h"
#include "fit.h"
#include "free.h"
#include "init.h"
#include "io.h"
#include "rio.h"
#include "matLib.h"
#include "optimize.h"
#include "precalc.h"

void ccosmo(
	   int *seqs,
	   int *numSeqs,
	   int *maxSeqLength,
	   char **constraints,
	   int *minWidth,
	   int *maxWidth,
	   int *revComp,
	   int *minSites,
	   int *maxSites,
	   int *oops,
	   int *zoops,
	   int *tcm,
	   char **approx,
	   int *cutFac,
	   char **wCrit,
	   int *wFold,
	   int *wTrunc,
	   char **modCrit,
	   int *modFold,
	   int *modTrunc,
	   char **conCrit,
	   int *conFold,
	   int *conTrunc,
	   char **intCrit,
	   int *intFold,
	   int *intTrunc,
	   int *starts,
	   int *lstarts,
	   int *backSeqs,
	   int *numBackSeqs,
	   int *maxBackSeqLength,
	   int *backFold,
	   double *transMat,
	   int *order,
	   int *selOrder,
	   double *klDivs,
	   double *estTransMat,
	   double *wCrits,
	   double *modCrits,
	   double *conCrits,
	   int *selWidth,	   
	   int *selModel,
	   int *selConSet,
	   int *selNumSites,
	   int *numModels,
	   double *pwm,
	   double *motProbs,
	   int *motRevComp,
	   int *alignSeqs,
	   int *alignStarts,
	   double *eVal,
	   int *success
	   )
{
  
	  
  DATASET dataSet;     /* the dataset storing all relevant information */


  *success = 1;
  
  /* set default values for dataset */
  if (!setDefaults(&dataSet)){
    (*success) = 0;
    return;
  }

  dataSet.R                                             = TRUE;
  dataSet.numSeqs                                       = *numSeqs;
  dataSet.maxSeqLength                                  = *maxSeqLength;
  dataSet.minWidth                                      = *minWidth;                 /* minimum motif width to consider */
  dataSet.maxWidth                                      = *maxWidth;                 /* maximum motif width to consider */
  dataSet.oops                                          = *oops;                     /* BOOLEAN: consider OOPS model? */
  dataSet.zoops                                         = *zoops;                    /* BOOLEAN: consider ZOOPS model? */
  dataSet.tcm                                           = *tcm;                      /* BOOLEAN: consider TCM model? */
  dataSet.revComp                                       = *revComp;                  /* BOOLEAN: consider reverse complement? */

  if ((minSites)!=NULL)       dataSet.minNumSites       = *minSites;              /* minimum number of motif sites */
  if ((maxSites)!=NULL)       dataSet.maxNumSites       = *maxSites;              /* maximum number of motif sites */
  dataSet.cutFactor                                     = *cutFac;
  dataSet.wCrit.fold                                    = *wFold;
  dataSet.wCrit.quant                                   = ((double) *wTrunc)/100;
  dataSet.wCrit.trunc                                   = dataSet.wCrit.quant!=1.0;
  dataSet.modCrit.fold                                  = *modFold;
  dataSet.modCrit.quant                                 = ((double) *modTrunc)/100;
  dataSet.modCrit.trunc                                 = dataSet.modCrit.quant!=1.0;
  dataSet.conCrit.fold                                  = *conFold;
  dataSet.conCrit.quant                                 = ((double) *conTrunc)/100;
  dataSet.conCrit.trunc                                 = dataSet.conCrit.quant!=1.0;
  dataSet.intCrit.fold                                  = *intFold;
  dataSet.intCrit.quant                                 = ((double)*intTrunc)/100;
  dataSet.intCrit.trunc                                 = dataSet.intCrit.quant!=1.0;
  dataSet.numStarts                                     = *starts;
  dataSet.likStarts                                     = *lstarts;
  dataSet.numBackSeqs                                   = *numBackSeqs;
  dataSet.maxBackSeqLength                              = *maxBackSeqLength;
  dataSet.backFold                                      = *backFold;

  if (order!=NULL){
                              dataSet.order             = *order;
                              dataSet.orderByCV         = FALSE;
  }

  if (!strcmp(*approx,"exact")){
    dataSet.approx = TCM;
  }else if (!strcmp(*approx,"over")){
    dataSet.approx = OVER;
  }else if (!strcmp(*approx,"cut")){
    dataSet.approx = CUT;
  }else{
    Rprintf("approx must be one of: exact, over, cut. Got: %s\n", *approx);
    (*success) = 0;
    return;
  }

  if (!strcmp(*wCrit,"lik")){
    dataSet.wCrit.crit = lik;
  }else if (!strcmp(*wCrit,"aic")){
    dataSet.wCrit.crit = aic;
  }else if (!strcmp(*wCrit,"bic")){
    dataSet.wCrit.crit = bic;
  }else if (!strcmp(*wCrit,"eval")){
    dataSet.wCrit.crit = eval;
  }else if (!strcmp(*wCrit,"likCV")){
    dataSet.wCrit.crit = likCV;
  }else{
    Rprintf("wCrit must be one of: lik, aic, bic, eval, likCV. Got: %s\n", *wCrit);
    (*success) = 0;
    return;
  }

  if (!strcmp(*modCrit,"lik")){
    dataSet.modCrit.crit = lik;
  }else if (!strcmp(*modCrit,"aic")){
    dataSet.modCrit.crit = aic;
  }else if (!strcmp(*modCrit,"bic")){
    dataSet.modCrit.crit = bic;
  }else if (!strcmp(*modCrit,"eval")){
    dataSet.modCrit.crit = eval;
  }else if (!strcmp(*modCrit,"likCV")){
    dataSet.modCrit.crit = likCV;
  }else{
    Rprintf("modCrit must be one of: lik, aic, bic, eval, likCV. Got: %s\n", *modCrit);
    (*success) = 0;
    return;
  }

  if (!strcmp(*conCrit,"lik")){
    dataSet.conCrit.crit = lik;
  }else if (!strcmp(*conCrit,"aic")){
    dataSet.conCrit.crit = aic;
  }else if (!strcmp(*conCrit,"bic")){
    dataSet.conCrit.crit = bic;
  }else if (!strcmp(*conCrit,"eval")){
    dataSet.conCrit.crit = eval;
  }else if (!strcmp(*conCrit,"likCV")){
    dataSet.conCrit.crit = likCV;
  }else if (!strcmp(*conCrit,"pwmCV")){
    dataSet.conCrit.crit = pwmCV;
  }else{
    Rprintf("conCrit must be one of: lik, aic, bic, eval, likCV, pwmCV. Got: %s\n", *conCrit);
    (*success) = 0;
    return;
  }

  if (!strcmp(*intCrit,"lik")){
    dataSet.intCrit.crit = lik;
  }else if (!strcmp(*intCrit,"eval")){
    dataSet.intCrit.crit = eval;
  }else if (!strcmp(*intCrit,"likCV")){
    dataSet.intCrit.crit = likCV;
  }else{
    Rprintf("intCrit must be one of: lik, eval, likCV. Got: %s\n", *intCrit);
    (*success) = 0;
    return;
  }
   
    
  /* process command line arguments */
  if (!processArgs(&dataSet)){
    (*success) = 0;
    return;
  }

  /* open stream to constraint string */
  if (constraints!=NULL){
    if((dataSet.conFile = fmemopen(*constraints,strlen(*constraints),"r")) == NULL){
      Rprintf("Can't open constraint stream.\n");
      return;
    }
    dataSet.hasConstraints = TRUE;
  }
 
 
  /* read in sequences */
  if (!readSeqsFromVec(seqs, &dataSet)){
    (*success) = 0;
    return;
  }

  /* read in background sequences if necessary */
  if (backSeqs!=NULL){
    if (!readBackSeqsFromVec(backSeqs, &dataSet)){
      (*success) = 0;
      return;
    }
  }

  /* get transition matrix */
  if (transMat!=NULL){
    if (!readTransMatFromVec(transMat,&dataSet)){
      (*success) = 0;
      return;
    }
  }else{
    if (!getTransMat(dataSet.orderByCV, &dataSet)){
      (*success) = 0;
      return;
    }
  }

  /* read in constraints */
  if (!readConFile(&dataSet)){
    (*success) = 0;
    return;
  }

  /* interpret constraints for different widths etc. */
  if (!deriveAllModels(&dataSet)){
    (*success) = 0;
    return;
  }

  /* create cut sequences if necessary */ 
  if (dataSet.approx!=TCM){
    if (!cutSequences(&dataSet)){
      (*success) = 0;
      return;
    }
  }

  /* print first status message if desired */
  if (dataSet.printStatus) Rprintf("0/%d   0/%d\n", dataSet.numSeqs, dataSet.effNumModels);


  /* precalculate arrays for background model */
  if (!precalcBack(&dataSet)){
    (*success) = 0;
    return;
  }

  /* get starting values */
  if (!getStarts(&dataSet)){
    (*success) = 0;
    return;
  }

  /* get pointer to dataSet to NLP namespace */
  initNLP(&dataSet);
  
  /* select model parameters */
  selectModel(&dataSet);

  /* fit final model */
  if (!finalModel(&dataSet)){
    Rprintf( "Final model did not converge. Are constraints feasible?\n");
    (*success) = 0;
    return;
  }

  /* generate output */
  output2vec(&dataSet, selOrder, klDivs, estTransMat, wCrits, modCrits, conCrits, selWidth, selModel, selConSet, selNumSites, numModels, pwm, motProbs, motRevComp, alignSeqs, alignStarts, eVal);

}

