/*****************************************************************************
*	   		     					             *
*      cosmo - Constrained search for motifs in DNA sequences                *
*      Copyright (C) 2006  Oliver Bembom (bembom@berkeley.edu)               *
*                                                                            *
*      This file is part of cosmo.                                           *
*                                                                            *
*      cosmo is free software; you can redistribute it and/or                *
*      modify it under the terms of the GNU Lesser General Public            *
*      License as published by the Free Software Foundation; either          *
*      version 2.1 of the License, or (at your option) any later version.    *
*                                                                            *
*      cosmo is distributed in the hope that it will be useful,              *
*      but WITHOUT ANY WARRANTY; without even the implied warranty of        *
*      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU     *
*      Lesser General Public License for more details.                       *
*                                                                            *
*      You should have received a copy of the GNU Lesser General Public      *
*      License along with this program; if not, write to the Free Software   *
*      Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,            *
*      MA  02110-1301, USA.                                                  *
*                                                                            *
*                                                                            *
*****************************************************************************/


/*****************************************************************************
*                                                                            *
*      This module contains functions for finding starting values based on   *
*      the E-value criterion.                                                *                                     
*                                                                            *
*****************************************************************************/

#include <R_ext/PrtUtil.h>
#include <R_ext/Memory.h>
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <float.h>
#include "macros.h"
#include "cosmo.h"
#include "alloc.h"
#include "estarts.h"
#include "eval.h"
#include "lstarts.h"
#include "matLib.h"
#include "posterior.h"
#include "theta.h"
  
  

/**********************************************************************/
/*
  getStarts

  Get starting values, either based on the likelihood function or 
  based on the E-value criterion.
  Each model type has its own starting value space at
  
  dataSet->svSpace + modTypeIndex

  since each model type may use a different number of starting values.

  On exit, the starting values have been written into the appropriate
  starting value spaces.

*/
/**********************************************************************/
int getStarts(
	       DATASET *dataSet  /* the dataset */
	       )
{

  /* initialize starting value space */
  if (!initSvSpace(dataSet)) return(0);
  
  if (dataSet->likStarts){
    /* initialize svLogLR arrays for starting values */
    if (!getSvLogLR(dataSet)) return(0);
    
    /* extract starting values with highest log-likelihoods */
    if (!xLikStarts(dataSet)) return(0);

  }else{
    if (!getEvalStarts(dataSet)) return(0);
  }

  /* add additional necessary information to starting values */
  if (!completeSv(dataSet)) return(0);


  return(1);
}

/**********************************************************************/
/*
  initSvSpace

  Initialize starting value spaces and allocate memory accordingly.


*/
/**********************************************************************/
int initSvSpace(
		 DATASET *dataSet  /* the dataset */
		 )
{
  int minWidth;               /* minimum width to consider */
  int maxWidth;               /* maximum width to consider */
  int maxSeqLength;           /* maximum sequence length */
  int numSeqs;                /* number of sequences */
  int nSiteNum;               /* index for looping over candidate values for the expected number of motif occurrences */
  int nSites;                 /* canidate value for the expected number of motif occurrences */
  int myNumSites;             /* index for looping over these candidate values */
  int minNumSites;            /* minimum of these candidate values */
  int maxNumSites;            /* maximum of these candidate values */
  int maxNumProfParms;        /* maximum number of IC profile parameters */
  int numConSets;             /* number of constraint sets */
  int totSv;                  /* total number of starting values */
  SVSPACE *svSpace;           /* starting value space */


  numConSets      = dataSet->numConSets;
  minWidth        = dataSet->minWidth;
  maxWidth        = dataSet->maxWidth;
  maxSeqLength    = dataSet->maxSeqLength;
  numSeqs         = dataSet->numSeqs;
  maxNumProfParms = dataSet->maxNumProfParms;

  dataSet->svSpace = svSpaceMalloc(dataSet->numModTypes);
  
  /* starting value space for OOPS model */
  if (dataSet->oops){
    svSpace                                = dataSet->svSpace;
    svSpace->mType                         = OOPS;

    svSpace->nnSites                       = 1;
    svSpace->numSites                      = IntMalloc(svSpace->nnSites);
    svSpace->numSites[0]                   = numSeqs;

    svSpace->numSv                         = dataSet->numStarts;

    totSv                                  = (maxWidth - minWidth + 1)*numConSets*svSpace->nnSites*svSpace->numSv;
    svSpace->totSv                         = totSv;
    svSpace->svTheta                       = thetaMalloc(totSv, maxWidth, maxNumProfParms);
    svSpace->svLogLR                       = DoubleMalloc((maxWidth - minWidth + 1)*numSeqs*maxSeqLength*svSpace->nnSites);    

    if (DEBUG>=3){
      Rprintf("initStartVals: Setting up dataSet->oopsSvspace\n");
      Rprintf("totSv: %d\n", totSv);
    }
  }

  /* starting value space for ZOOPS model */
  if (dataSet->zoops){

    svSpace                                = dataSet->svSpace + (int) dataSet->oops;
    svSpace->mType                         = ZOOPS;

    minNumSites                            = MIN(dataSet->minNumSites,numSeqs);
    maxNumSites                            = MIN(dataSet->maxNumSites,numSeqs);

    for (nSites=minNumSites, svSpace->nnSites=0; nSites < 2*maxNumSites; nSites *= 2, (svSpace->nnSites)++);

    svSpace->numSites                      = IntMalloc(svSpace->nnSites);

    svSpace->numSv                         = dataSet->numStarts;

    totSv                                  = (maxWidth - minWidth + 1)*numConSets*svSpace->nnSites*svSpace->numSv;
    svSpace->totSv                         = totSv;
    svSpace->svTheta                       = thetaMalloc(totSv, maxWidth, maxNumProfParms);
    svSpace->svLogLR                       = DoubleMalloc((maxWidth - minWidth + 1)*numSeqs*maxSeqLength*svSpace->nnSites);    

    for (nSiteNum=0, nSites = minNumSites;nSiteNum<svSpace->nnSites;nSiteNum++, nSites *=2){
      svSpace->numSites[nSiteNum] = nSites > maxNumSites ? maxNumSites : nSites;
    }


    if (DEBUG>=3){
      Rprintf("initStartVals: Setting up dataSet->zoopsSvspace\n");
      Rprintf("totSv: %d\n", totSv);

      for (nSiteNum=0;nSiteNum<svSpace->nnSites;nSiteNum++){
	Rprintf("Number of sites: %d\n", svSpace->numSites[nSiteNum]);
      }	
    }
  }

  /* starting value space for tcm model */
  if (dataSet->tcm){

    svSpace                                = dataSet->svSpace + (int) dataSet->oops + (int) dataSet->zoops;
    svSpace->mType                         = TCM;

    minNumSites                            = dataSet->minNumSites;
    maxNumSites                            = dataSet->maxNumSites;

    for (myNumSites=minNumSites, nSiteNum=0;myNumSites<2*maxNumSites;myNumSites *= 2) nSiteNum ++;

    svSpace->nnSites                       = nSiteNum;
    svSpace->numSv                         = dataSet->numStarts;

    svSpace->numSites                      = IntMalloc(svSpace->nnSites);


    totSv                                  = (maxWidth - minWidth + 1)*numConSets*svSpace->nnSites*svSpace->numSv;
    svSpace->totSv                         = totSv;
    svSpace->svTheta                       = thetaMalloc(totSv, maxWidth, maxNumProfParms);
    svSpace->svLogLR                       = DoubleMalloc((maxWidth - minWidth + 1)*numSeqs*maxSeqLength*svSpace->nnSites);    

    for (myNumSites=minNumSites, nSiteNum=0;myNumSites<2*maxNumSites;myNumSites *= 2, nSiteNum++){
      svSpace->numSites[nSiteNum] = (myNumSites>maxNumSites) ? maxNumSites : (int) myNumSites;
    }

    if (DEBUG>=3){
      Rprintf("initStartVals: Setting up dataSet->tcmSvspace\n");
      Rprintf("totSv: %d\n", totSv);

      for (nSiteNum=0;nSiteNum<svSpace->nnSites;nSiteNum++){
	Rprintf("Number of sites: %d\n", svSpace->numSites[nSiteNum]);
      }	
    }
  }

  /* get maximum number of nSites */
  if (!getMaxNnSites(dataSet)) return(0);

  return(1);
}


/**********************************************************************/
/*
  getMaxNnSites

  Get maximum of number of candidates for the expected number of 
  motif occurrences over all model types.

  Assigns that maximum to dataSet->maxNnSites.


*/
/**********************************************************************/
int getMaxNnSites(
		   DATASET *dataSet   /* the dataset */
		   )
{

  int modTypeNum;   /* index for looping over model types */

  dataSet->maxNnSites = 0;
  for (modTypeNum=0; modTypeNum<dataSet->numModTypes; modTypeNum++){
    dataSet->maxNnSites = MAX(dataSet->maxNnSites, dataSet->svSpace[modTypeNum].nnSites);
  }

  return(1);
}

/**********************************************************************/
/*
  satisfiesConstraints

  Check if a candidate starting value satisfies the following constraints:
  
    - position frequency constraints that give lower bounds on the 
      nucleotide frequency at a particular position of the motif
    - interval frequency constraints that give lower bounds on the
      average nucleotide frequency across a whole interval

  Returns TRUE if starting value satisfies these constraints, 
  FALSE otherwise.


*/
/**********************************************************************/
BOOLEAN satisfiesConstraints(
			     double *pwm,   /* the PWM of the candidate starting value */
			     int width,     /* the width of this PWM */
			     MODEL *mod     /* the interpreted constraint set to check against */
			     )
{

  BOOLEAN isGoodStartVal;   /* does this starting value satisfy the relevant constraints? */

  isGoodStartVal = TRUE;

  /* Check if FreqCon and PosCon constraints are fulfilled; else discard */
  if (mod->hasPosFreqCons || mod->numLinFreqCons>0){
    
    /* check bounds on PWM parameters */
    if (mod->hasPosFreqCons && !checkPWMBounds(pwm, width, mod)) isGoodStartVal = FALSE;
    
    /* check linear frequency constraints */
    if (!checkLinFreqCons(pwm, width, mod)) isGoodStartVal = FALSE;
    
  }

  return isGoodStartVal;
}


/**********************************************************************/
/*
  checkPWMBounds

  Check if a candidate starting value satisfies the position frequency
  constraints of a given constraint set that give lower bounds for
  entries of the position weight matrix.

  Returns TRUE if starting value satisfies these constraints, 
  FALSE otherwise.


*/
/**********************************************************************/
BOOLEAN checkPWMBounds(
		       double *pwm,   /* PWM of candidate starting value */
		       int width,     /* width of this PWM */
		       MODEL *mod     /* interpreted constraint set */
		       )
{

  int pos;               /* index for looping over positions of PWM */ 
  BOOLEAN areSatisfied;  /* are constraints satisfied? */

  areSatisfied = TRUE;
  for (pos=0;pos<4*width;pos++){
    if (pwm[pos]>mod->up[pos] || pwm[pos]<mod->low[pos]) areSatisfied = FALSE;
  }

  return areSatisfied;
}


/**********************************************************************/
/*
  checkLinFreqCons

  Check if a candidate starting value satisfies the interval frequency
  constraints of a given constraint set that give lower bounds for
  linear combinations of entries of the position weight matrix.

  Returns TRUE if starting value satisfies these constraints, 
  FALSE otherwise.


*/
/**********************************************************************/
BOOLEAN checkLinFreqCons(
			 double *pwm,   /* PWM of candidate starting value */
			 int width,     /* width of this PWM */
			 MODEL *mod     /* interpreted constraint set */
			 )
{

  int pos;               /* index for looping over positions of PWM */
  int conNum;            /* index for looping over linear constraints */
  BOOLEAN areSatisfied;  /* are constraints satisfied? */
  double sum;            /* sum of relevant entries of the PWM */

  areSatisfied = TRUE;
  
  for (conNum=0;conNum<mod->numLinFreqCons;conNum++){
    sum = 0;
    for (pos=0;pos<width*4;pos++){
      /* linear frequency constraints start in column 'width' of linConMatrix */
      sum += pwm[pos]*mod->linConMatrix[LCM(width+conNum,pos)];
    }
    /* bounds on linear frequency constraints start at position mod->n + width in mod->low and mod->up */
    if (sum<mod->low[mod->n+width + conNum] || sum>mod->up[mod->n+width + conNum]) areSatisfied = FALSE;
  }

  return areSatisfied;
}


/**********************************************************************/
/*
  seq2logPwm
  
  Convert an oligo into a position weight matrix for a candidate 
  starting value by assigning the observed nucleotides a probability
  'hiProb' and the remaining ones the probability (1-hiProb)/3
  for all positions with regular characters, and assigning probability
  0.25 to all nucleotides in positions with unknown characters 'X'.

  The position weight matrix is log-transformed and stored in integer
  format by calling the macro INT_LOG.

  The position weight matrix is written to lpwm.


*/
/**********************************************************************/
int seq2logPwm(
		int *lpwm,          /* on exit: the log-transformed PWM */
		int *seq,           /* the sequence from which to construct the PWM */
		int width,          /* the width of the motif */
		double hiProb       /* the probability to assign to observed nucleotides */
		)
{

  double lowProb;    /* the probability to assign to unobserved nucleotides */
  int pos;           /* index for looping over positions of motif */ 
  int nuc;           /* index for looping over nucleotides */

  lowProb = (1-hiProb)/3;

  for (pos=0;pos<width;pos++){
    if (seq[pos]<4){
      /* regular character */
      for (nuc=0;nuc<4;nuc++){
	lpwm[PWM(pos,nuc)] = INT_LOG(lowProb);
      }
      lpwm[PWM(pos,seq[pos])] = INT_LOG(hiProb);
    }else{
      /* unknown character 'X' */
      for (nuc=0;nuc<4;nuc++) lpwm[PWM(pos,nuc)] = INT_LOG(1.0/4);
    }
  }

  return(1);
} 

/**********************************************************************/
/*
  addCol2lPwm
  
  Add another column to a log-transformed position weight matrix.

 
*/
/**********************************************************************/
int addCol2lPwm(
		 int *lPwm,     /* the PWM to which we are adding a column */
		 int *seq,      /* the sequence from which the PWM was constructed */
		 int start,     /* the starting position of the corresponding oligo */
		 int newWidth,  /* the new width of the PWM */
		 double hiProb  /* the probability to assign to observed nucleotides */
		 )
{

  double lowProb;    /* probability to assign to unobserved nucleotides */
  int lastCol;       /* last column of motif */
  int nuc;           /* nucleotide */

  lowProb = (1-hiProb)/3;

  /* add new column to PWM */
  lastCol = newWidth - 1;
  if (seq[start+lastCol]<4){
    /* regular characters */
    for (nuc=0;nuc<4;nuc++){
      lPwm[PWM(lastCol,nuc)] = INT_LOG(lowProb);
    }
    lPwm[PWM(lastCol,seq[start+lastCol])] = INT_LOG(hiProb);
  }else{
    /* unknown character 'X' */
    for (nuc=0;nuc<4;nuc++) lPwm[PWM(lastCol,nuc)] = INT_LOG(1.0/4);
  }
  return(1);
}


/**********************************************************************/
/*
  getEvalStarts
  
  Find starting values based on the E-value criterion.

  Starting are written to appropriate starting value spaces in the
  dataset.

  Note: Starting values are not yet complete, i.e. they are still 
  missing some information. Call completeSv() for this purpose.


*/
/**********************************************************************/
int getEvalStarts(
		   DATASET *dataSet  /* the dataset */
		   )
{

  int *lPwm;           /* space for a log-transformed PWM */
  int *prevLPwm;       /* space for another log-transformed PWM */
  double hiProb;       /* probability to assign to observed nucleotides in constructing PWM from oligo */

  int width;           /* index for looping over candidate motif widths */
  int minWidth;        /* minimum motif width to consider */
  int maxWidth;        /* maximum motif width to consider */
  BOOLEAN oneWidth;    /* is there only one candidate width? */

  int nnSites;         /* number of candidate values for the expected number of motif occurrences */
  int numSites;        /* candidate value for the expected number of motif occurrences */
  int nSitesNum;       /* index of candidate value */

  int oligoSeqNum;     /* number of sequence from which oligo is taken */
  int oligoStart;      /* position in sequence at which oligo starts */
  int *oligoSeq;       /* oligo sequence */
  long oligoSeqLength; /* length of oligo oligo sequence */

  int numSeqs;         /* number of sequences */
  MTYPE mType;         /* model type */
  int modTypeNum;      /* index of candidate model type */
  int numModTypes;     /* number of canidate model types */
  int numConSets;      /* number of constraint sets */
  int conSetNum;       /* index for looping over constraint sets */

  int numSv;           /* number of starting values to be found for given expected number of motif occurrences */
  int startNum;        /* index for looping over different such starting values */
  THETA *svTheta;      /* one starting value */
  SVSPACE *svSpace;    /* space of starting values */

  long offset;         /* offset to starting values for current width, model type, constraint set, and intensity parameter */



  minWidth       = dataSet->minWidth;
  maxWidth       = dataSet->maxWidth;
  lPwm           = dataSet->lPwm1;
  prevLPwm       = dataSet->lPwm2;
  hiProb         = dataSet->hiProb;
  oneWidth       = (dataSet->maxWidth==dataSet->minWidth);
  numSeqs        = dataSet->numSeqs;
  numModTypes    = dataSet->numModTypes;
  numConSets     = dataSet->numConSets;


 
  if (DEBUG>=2) Rprintf("\n");
  for (oligoSeqNum=0;oligoSeqNum<numSeqs;oligoSeqNum++){
    if(DEBUG>=2) Rprintf( "\reGetStart: Extracting starting values from sequence %d/%d", oligoSeqNum+1, numSeqs);

    /* print first status message if desired */
    if (dataSet->printStatus) Rprintf("%d/%d   0/%d\n", oligoSeqNum+1, dataSet->numSeqs, dataSet->effNumModels);

   
    oligoSeq       = dataSet->samples[oligoSeqNum].seq;
    oligoSeqLength = dataSet->samples[oligoSeqNum].length;

    /* loop over start positions for oligo */
    for (oligoStart=0;oligoStart<oligoSeqLength-minWidth + 1; oligoStart++){

      /* set up with width = minWidth*/
      width = minWidth;

      /* Create pwm from new oligo */
      seq2logPwm(lPwm, oligoSeq + oligoStart, width, hiProb);

      if (oligoStart!=0){
	/* Create pwm from previous oligo */
	seq2logPwm(prevLPwm, oligoSeq + oligoStart - 1, width, hiProb);

	/* calculate mProbs1 from previous mProbs*/
	if (!eGetOverMProbs(dataSet, lPwm, prevLPwm, width, 1,oneWidth)) return(0);

      }else{
	/* calculate mProbs1 fresh*/
	if (!eGetNewMProbs(dataSet, lPwm, width, 1, FALSE, NULL)) return(0);
      }

      /* check out the starting values obtained from this oligo */
      if (!checkStart(dataSet,1,width)) return(0);


      /* copy mProbs1 to mProbs2*/
      if (!oneWidth){
	if (!copyIntMotProbs(dataSet->intMotProbs1, dataSet->intMotProbs2, dataSet)) return(0);
      }

      /* fill in for oligoStart = 0 and minWidth<width<=maxWidth */
      /* be careful to respect oligoStart<oligoSeqLength-width+1 */
      for (width = minWidth+1;width<=MIN(maxWidth,oligoSeqLength-oligoStart);width++){
	  
	/* add column to pwm */
	addCol2lPwm(lPwm, oligoSeq, oligoStart, width, hiProb);

	/* update mProbs2 */
	if (!eGetHigherMProbs(dataSet, lPwm, width, 2)) return(0);


	/* copy mProbs2 to mProbs1 for getOverMProbs */
	if (width==minWidth+1){
	  if (!copyIntMotProbs(dataSet->intMotProbs2, dataSet->intMotProbs1, dataSet)) return(0);
	}

	/* check out the starting values obtained from this oligo */
	if (!checkStart(dataSet,2,width)) return(0);

      }
    }	    
  }

  if (DEBUG>=2) Rprintf("\n");

  if (DEBUG>=4){
    for (width=minWidth; width<=maxWidth; width++){

      Rprintf("width=%d\n", width);

      for (modTypeNum=0; modTypeNum<numModTypes; modTypeNum++){
	svSpace   = dataSet->svSpace + modTypeNum;
	nnSites = svSpace->nnSites;
	mType   = svSpace->mType;
	numSv   = svSpace->numSv;

	Rprintf("modType = %d\n", mType);

	for (nSitesNum=0; nSitesNum<nnSites; nSitesNum++){
	  numSites = svSpace->numSites[nSitesNum];

	  Rprintf("numSites = %d\n", numSites);
	  for (conSetNum=0; conSetNum<numConSets; conSetNum++){
	  
	    Rprintf("conSetNum = %d\n", conSetNum);

	    offset = SV(width,conSetNum,nSitesNum, 0);
	    svTheta = svSpace->svTheta + offset;
	    Rprintf("offset = %ld\n", offset);
	    
	    for (startNum=0; startNum<numSv; startNum++){
	      Rprintf("startvalue %d: score = %lf\n", startNum+1, svTheta[startNum].score);
	      PrintDoubleMatrix2File(stderr,svTheta[startNum].pwm, 4, width);
	    }
	  }
	}
      }
    }
  }

  return(1);
}


/**********************************************************************/
/*
  eGetNewMProbs
  
  Get log-transformed motif probabilities in integer format by
  calculation from scratch for a given PWM.

*/
/**********************************************************************/ 
int eGetNewMProbs(
		   DATASET *dataSet,    /* the dataset */
		   int *lPwm,           /* PWM for which to calculate motif probabilities */ 
		   int width,           /* width of this PWM */
		   int spaceNum,        /* store mProbs in intMotProbs1 or intMotProbs2 ? */
		   BOOLEAN excludeAny,  /* are any sequences to be excluded for CV purposes? */
		   int *excludeCv       /* array indicating which sequences are to be exclude, NULL if none are to be excluded */
		   )
{

  int seqNum;                   /* index of current sequence */
  int numSeqs;                  /* total number of sequences */
  int start;                    /* staring position of current motif site */
  int seqLength;                /* length of current sequence */
  int mProb;                    /* log-transformed motif probability */
  int pos;                      /* position within the motif */
  int *seq;                     /* current sequence */
  INTMOTPROB *intMotProbs;      /* array of motif probabilities, one for each possible motif site */
  INTMOTPROB *seqIntMotProbs;   /* pointer to these motif probabilities for current sequence */
  long maxSeqLength;            /* maximum sequence length */
  long motProbLength;           /* this may be twice as long if we need motif probabilities for the reverse complement */
  
  /* check that motif probabilities are to be stored either in mProbs1 or mProbs 2 */
  if (spaceNum!=1 && spaceNum!=2){
    Rprintf("egetNewMProbs: spaceNum must be 1 or 2. Got %d. Exiting ...\n", spaceNum);
    return(0);
  }

  numSeqs       = dataSet->numSeqs;
  maxSeqLength  = dataSet->maxSeqLength;
  motProbLength = maxSeqLength*(1 + dataSet->revComp);
  intMotProbs   = spaceNum==1 ? dataSet->intMotProbs1 : dataSet->intMotProbs2; 


  /* loop over sequences */
  for (seqNum=0;seqNum<numSeqs;seqNum++){

    /* exclude this sequence? */
    if (excludeAny && excludeCv[dataSet->samples[seqNum].cvGroup]) continue;

    seq            = dataSet->samples[seqNum].seq;
    seqLength      = dataSet->samples[seqNum].length;
    seqIntMotProbs = intMotProbs + seqNum*motProbLength; 

    /* loop over possible motif start sites */
    for (start=0;start<seqLength - width + 1;start++){
      mProb = 0;
      for (pos=0;pos<width;pos++){
	if (seq[start+pos] < 4){
	  mProb += lPwm[PWM(pos, seq[start+pos])];
	}else{
	  mProb += LOG(1.0/4);
	}
      }
      seqIntMotProbs[start].prob = mProb;
      seqIntMotProbs[start].seq  = seq + start;
    }
    
    /* fill in motif start sites at the end with small probabilities */
    for (start=seqLength-width+1; start<seqLength; start++){
      seqIntMotProbs[start].prob = SMALL;
      seqIntMotProbs[start].seq  = NULL;
    }      

    /* motif probabilities on reverse complememnt */
    if (dataSet->revComp){

      seq            = dataSet->samples[seqNum].rcSeq;
      seqIntMotProbs = intMotProbs + seqNum*motProbLength + maxSeqLength; 
      
      for (start=0;start<seqLength - width + 1;start++){
	mProb = 0;
	for (pos=0;pos<width;pos++){
	  if (seq[start+pos]<4){
	    mProb += lPwm[PWM(pos, seq[start+pos])];
	  }else{
	    mProb += LOG(1.0/4);
	  }
	}
	seqIntMotProbs[start].prob = mProb;
	seqIntMotProbs[start].seq  = seq + start;
      }

      for (start=seqLength-width+1; start<seqLength; start++){
	seqIntMotProbs[start].prob = SMALL;
	seqIntMotProbs[start].seq  = NULL;
      }      
    }
  }

  return(1);
}


/**********************************************************************/
/*
  eGetHigherMProbs
  
  Calculate log-transformed motif probabilities for width W from the 
  stored motif probabilities for width W-1, which are stored either
  in intMotProbs1 or intMotProbs2.

  New motif probabilities are stored in the same array.

*/
/**********************************************************************/
int eGetHigherMProbs(
		      DATASET *dataSet,    /* the dataset */
		      int *lPwm,           /* log-transformed PWM for which motif probabilities are desired */
		      int width,           /* widht of this PWM */
		      int spaceNum         /* are motif probs for width-1 stored in intMotProbs1 or intMotProbs2 ? */
		      )
{


  int seqNum;                   /* index of current sequence */
  int numSeqs;                  /* total number of sequences */
  int start;                    /* staring position of current motif site */
  int seqLength;                /* length of current sequence */
  int lastCol;                  /* last position in motif */
  int *seq;                     /* current sequence */
  INTMOTPROB *intMotProbs;      /* array of motif probabilities, one for each possible motif site */
  INTMOTPROB *seqIntMotProbs;   /* pointer to these motif probabilities for current sequence */
  long maxSeqLength;            /* maximum sequence length */
  long motProbLength;           /* this may be twice as long if we need motif probabilities for the reverse complement */


  if (spaceNum!=1 && spaceNum!=2){
    Rprintf("getHigherMProbs: spaceNum must be 1 or 2. Got %d. Exiting ...\n", spaceNum);
    return(0);
  }

  numSeqs       = dataSet->numSeqs;
  maxSeqLength  = dataSet->maxSeqLength;
  lastCol       = width - 1;
  motProbLength = maxSeqLength*(1 + dataSet->revComp);
  intMotProbs   = spaceNum==1 ? dataSet->intMotProbs1 : dataSet->intMotProbs2; 

  /* loop over sequences */
  for (seqNum=0; seqNum<numSeqs; seqNum++){

    seq       = dataSet->samples[seqNum].seq;
    seqLength = dataSet->samples[seqNum].length;
    seqIntMotProbs = intMotProbs + seqNum*motProbLength;
    
    /* get motif probabilities by simpling adding log(freq) of last column */
    for (start=0;start<seqLength - width + 1;start++){
      if (seq[start+lastCol]<4){
	seqIntMotProbs[start].prob += lPwm[PWM(lastCol,seq[start+lastCol])];
      }else{
	seqIntMotProbs[start].prob += LOG(1.0/4);
      }
      seqIntMotProbs[start].seq   = seq + start;
    }

    /* fill in remaining motif probabilities at end of sequence with small numbers */
    for (start=seqLength-width+1;start<seqLength;start++){
      seqIntMotProbs[start].prob  = SMALL;
      seqIntMotProbs[start].seq   = NULL;
    }

    /* find motif probabilities for reverse complement */
    if (dataSet->revComp){
      seqIntMotProbs = intMotProbs + seqNum*motProbLength + maxSeqLength;
      seq            = dataSet->samples[seqNum].rcSeq;

      /* get motif probabilities by simpling adding log(freq) of last column */
      for (start=0;start<seqLength - width + 1;start++){
	if (seq[start+lastCol]<4){
	  seqIntMotProbs[start].prob += lPwm[PWM(lastCol,seq[start+lastCol])];
	}else{
	  seqIntMotProbs[start].prob += LOG(1.0/4);
	}
	seqIntMotProbs[start].seq   = seq + start;
      }
      
      /* fill in remaining motif probabilities at end of sequence with small numbers */
      for (start=seqLength-width+1;start<seqLength;start++){
	seqIntMotProbs[start].prob  = SMALL;
	seqIntMotProbs[start].seq   = NULL;
      }
    }
  }

  return(1);
}


/**********************************************************************/
/*
  eGetOverMProbs
  
  Calculate log-transformed motif probabilities for width W and oligo
  at position l from the stored motif probabilities for width W+1 and 
  oligo at position l+1, which are stored either in intMotProbs1 or intMotProbs2.

  New motif probabilities are stored in the same array.

*/
/**********************************************************************/
int eGetOverMProbs(
		    DATASET *dataSet,  /* the dataset */
		    int *lPwm,         /* the log-transformed PWM for which to find the motif probabilities */
		    int *prevLPwm,     /* the log-transformed PWM for width+1 starting in the previous position */
		    int width,         /* the width of the current PWM */
		    int spaceNum,      /* are old motif probs stored in intMotProbs1 or intMotProbs2 ? */
		    BOOLEAN oneWidth   /* is there only one candidate width? */
		    )
{

  int seqNum;                   /* index of current sequence */
  int numSeqs;                  /* total number of sequences */
  int start;                    /* staring position of current motif site */
  int seqLength;                /* length of current sequence */
  int firstCol;                 /* first position in motif */
  int lastCol;                  /* last position in motif */
  int pos;                      /* index for looping over columns in motif */
  int *seq;                     /* current sequence */
  INTMOTPROB *intMotProbs;      /* array of motif probabilities, one for each possible motif site */
  INTMOTPROB *seqIntMotProbs;   /* pointer to these motif probabilities for current sequence */
  long maxSeqLength;            /* maximum sequence length */
  long motProbLength;           /* this may be twice as long if we need motif probabilities for the reverse complement */
  int mProb;                    /* log-transformed motif probability */



  if (spaceNum!=1 && spaceNum!=2){
    Rprintf("getHigherMProbs: spaceNum must be 1 or 2. Got %d. Exiting ...\n", spaceNum);
    return(0);
  }

  numSeqs       = dataSet->numSeqs;
  maxSeqLength  = dataSet->maxSeqLength;
  lastCol       = width - 1;
  firstCol      = 0;
  motProbLength = maxSeqLength*(1+dataSet->revComp);
  intMotProbs   = spaceNum==1 ? dataSet->intMotProbs1 : dataSet->intMotProbs2; 


    
  /* if there is only one candidate width, we have to use stored motif probs from same width */
  if (oneWidth){
    for (seqNum=0; seqNum<numSeqs; seqNum++){

      seq            = dataSet->samples[seqNum].seq;
      seqLength      = dataSet->samples[seqNum].length;
      seqIntMotProbs = intMotProbs + seqNum*motProbLength;

      /* get new motif probs by subtracting first column of old PWM and adding last column of new PWM */
      for (start = seqLength - width;start>0;start--){
	if ((seq[start+lastCol]<4) & (seq[start-1]<4)){
	  seqIntMotProbs[start].prob = seqIntMotProbs[start-1].prob + lPwm[PWM(lastCol,seq[start+lastCol])] -
	    prevLPwm[PWM(firstCol,seq[start-1])];
	}else if (seq[start+lastCol]<4){
	  seqIntMotProbs[start].prob = seqIntMotProbs[start-1].prob + lPwm[PWM(lastCol,seq[start+lastCol])] - LOG(1.0/4);
	}else if (seq[start-1]<4){
	  seqIntMotProbs[start].prob = seqIntMotProbs[start-1].prob + LOG(1.0/4) - prevLPwm[PWM(firstCol,seq[start-1])];
	}else{
	  seqIntMotProbs[start].prob = seqIntMotProbs[start-1].prob;
	}
	seqIntMotProbs[start].seq  = seq + start;
      }

      /* do start=0 last */
      start = 0;
      mProb = 0;
      for (pos=0;pos<width;pos++){
	if (seq[start+pos]<4){
	  mProb += lPwm[PWM(pos, seq[start+pos])];
	}else{
	  mProb += LOG(1.0/4);
	}
      }
      seqIntMotProbs[start].prob = mProb;
      seqIntMotProbs[start].seq  = seq + start;

      /* fill in remaining motif probabilities with small numbers */
      for (start=seqLength-width+1; start<seqLength; start++){
	seqIntMotProbs[start].prob = SMALL;
	seqIntMotProbs[start].seq  = NULL;
      }


      if (dataSet->revComp){
	seq     = dataSet->samples[seqNum].rcSeq;
	seqIntMotProbs = intMotProbs + seqNum*motProbLength + maxSeqLength;

	/* get new motif probs by subtracting first column of old PWM and adding last column of new PWM */
	for (start = seqLength - width;start>0;start--){
	  if ((seq[start+lastCol]<4) & (seq[start-1]<4)){
	    seqIntMotProbs[start].prob = seqIntMotProbs[start-1].prob + lPwm[PWM(lastCol,seq[start+lastCol])] -
	      prevLPwm[PWM(firstCol,seq[start-1])];
	  }else if (seq[start+lastCol]<4){
	    seqIntMotProbs[start].prob = seqIntMotProbs[start-1].prob + lPwm[PWM(lastCol,seq[start+lastCol])] - LOG(1.0/4);
	  }else if (seq[start-1]<4){
	    seqIntMotProbs[start].prob = seqIntMotProbs[start-1].prob + LOG(1.0/4) - prevLPwm[PWM(firstCol,seq[start-1])];
	  }else{
	  seqIntMotProbs[start].prob = seqIntMotProbs[start-1].prob;
	  }

	  seqIntMotProbs[start].seq  = seq + start;
	}

	/* do start=0 last */
	start = 0;
	mProb = 0;
	for (pos=0;pos<width;pos++){
	  if (seq[start+pos]<4){
	    mProb += lPwm[PWM(pos, seq[start + pos])];
	  }else{
	    mProb += LOG(1.0/4);
	  }
	}
	seqIntMotProbs[start].prob = mProb;
	seqIntMotProbs[start].seq  = seq + start;
	
	/* fill in remaining motif probabilities with small numbers */
	for (start=seqLength-width+1; start<seqLength; start++){
	  seqIntMotProbs[start].prob = SMALL;
	  seqIntMotProbs[start].seq  = NULL;
	}

      }
    }
  }else{

    /* if there is more than one candidate width, we can use motif probs stored for width+1 */
    for (seqNum=0; seqNum<numSeqs; seqNum++){

      seq            = dataSet->samples[seqNum].seq;
      seqLength      = dataSet->samples[seqNum].length;
      seqIntMotProbs = intMotProbs + seqNum*motProbLength;

      /* get new motif probs by subtracting contribution of first column of old PWM */
      for (start = seqLength - width;start>0;start--){
	if (seq[start-1]<4){
	  seqIntMotProbs[start].prob = seqIntMotProbs[start-1].prob - prevLPwm[PWM(firstCol,seq[start-1])];
	}else{
	  seqIntMotProbs[start].prob = seqIntMotProbs[start-1].prob - LOG(1.0/4);
	}
	seqIntMotProbs[start].seq  = seq + start;
      }

      /* do start=0 last */
      start = 0;
      mProb = 0;
      for (pos=0;pos<width;pos++){
	if (seq[start+pos]<4){
	  mProb += lPwm[PWM(pos, seq[start+pos])];
	}else{
	  mProb += LOG(1.0/4);
	}
      }
      seqIntMotProbs[start].prob = mProb;
      seqIntMotProbs[start].seq  = seq + start;

      /* fill in last entries with small probabilities */
      for (start=seqLength-width+1; start<seqLength; start++){
	seqIntMotProbs[start].prob = SMALL;
	seqIntMotProbs[start].seq  = NULL;
      }


      /* store probability of motif in opposite orientation, starting in same position */
      if (dataSet->revComp){
	seq            = dataSet->samples[seqNum].rcSeq;
	seqIntMotProbs = intMotProbs + seqNum*motProbLength + maxSeqLength;

	/* get new motif probs by subtracting contribution of first column of old PWM */
	for (start = seqLength - width;start>0;start--){
	  if (seq[start-1]<4){
	    seqIntMotProbs[start].prob = seqIntMotProbs[start-1].prob - prevLPwm[PWM(firstCol,seq[start-1])];
	  }else{
	    seqIntMotProbs[start].prob = seqIntMotProbs[start-1].prob - LOG(1.0/4);
	  }
	  seqIntMotProbs[start].seq  = seq + start;	  
	}

	/* do start=0 last */
	start = 0;
	mProb = 0;
	for (pos=0;pos<width;pos++){
	  if (seq[start+pos]<4){
	    mProb += lPwm[PWM(pos, seq[start + pos])];
	  }else{
	    mProb += LOG(1.0/4);
	  }
	}
	seqIntMotProbs[start].prob = mProb;
	seqIntMotProbs[start].seq  = seq + start;
	
	/* fill in last entries with small probabilities */
	for (start=seqLength-width+1; start<seqLength; start++){
	  seqIntMotProbs[start].prob = SMALL;
	  seqIntMotProbs[start].seq  = NULL;
	}
      }
    }
  }

  return(1);
}

/**********************************************************************/
/*
  copyIntMotProbs
  
  Copy log-transformed motif probabilities from mProbs1 to mProbs2 or
  the other way around

*/
/**********************************************************************/
int copyIntMotProbs(
		     INTMOTPROB *from,   /* copy these motif probabilities */
		     INTMOTPROB *to,     /* into this array */
		     DATASET *dataSet    /* the data set */
		     )
{
 
  int seqNum;          /* current sequence number */
  int start;           /* current position in sequence */
  int numSeqs;         /* number of sequences */
  INTMOTPROB *seqFrom; /* motif probabilities for current sequence that are to be copied */
  INTMOTPROB  *seqTo;  /* pointer to array of motif probabilities for current sequences */
  long maxSeqLength;   /* maximum sequence length */
  long motProbLength;  /* this can be twice as long as maxSeqLength if we consider the reverse complement */


  maxSeqLength  = dataSet->maxSeqLength;
  motProbLength = maxSeqLength*(1 + dataSet->revComp);
  numSeqs       = dataSet->numSeqs;

  for (seqNum=0; seqNum<numSeqs; seqNum++){
    seqFrom = from + seqNum*motProbLength;
    seqTo   = to   + seqNum*motProbLength;

    for (start=0; start<motProbLength; start++){
      seqTo[start].prob = seqFrom[start].prob;
      seqTo[start].seq  = seqFrom[start].seq;
    }
  }

  return(1);
}


/**********************************************************************/
/*
  compIntMotProbs
  
  Compare two log-transformed integer-valued motif probabilites.

  Returns -1 if first motif probability is greater,
           1 if first motif probability is less,
	   0 otherwise.
 
*/
/**********************************************************************/
int compIntMotProbs(
		    const void *x,   /* pointer to first motif probability */
		    const void *y    /* pointer to second motif probability */
		    )
{

  const INTMOTPROB *i1;
  const INTMOTPROB *i2;

  i1 = (const INTMOTPROB *) x;
  i2 = (const INTMOTPROB *) y;

  if (i1->prob > i2->prob) return -1;
  if (i2->prob > i1->prob) return 1;
  return 0;
}


/**********************************************************************/
/*
  sortIntMotProbs
  
  Sort the two log-transformed integer-valued motif probabilites and 
  find maximam
  If global=TRUE, all motif proababilites are sorted, under the constraint
  that no consecutive W nucleotides may contain two maxima.
  If global=FALSE, the maximum motif probability of each sequence is
  found.

  Returns the number of maximam found.
 
*/
/**********************************************************************/
int sortIntMotProbs(
		    DATASET *dataSet,   /* the dataset */
		    int spaceNum,       /* are motif probs stored in intMotProbs1 or intMotProbs2? */
		    BOOLEAN global,     /* global sort or within sequences? */
		    int width           /* width of motif */
		    )
{

  int seqNum;                   /* index of current sequence */
  int numSeqs;                  /* total number of sequences */
  int seqLength;                /* length of current sequence */

  int pos;                      /* index for looping over columns in motif */
  int *seq;                     /* current sequence */
  INTMOTPROB *intMotProbs;      /* array of motif probabilities, one for each possible motif site */
  INTMOTPROB *maxMotProbs;      /* maximum motif probabilities */
  long maxSeqLength;            /* maximum sequence length */
  long motProbLength;           /* this may be twice as long if we need motif probabilities for the reverse complement */

  int prob;                     /* current probability */
  int maxProb;                  /* current maximum probability */
  int lastPos;                  /* last position in sequence */
  int nextPos;                  /* next eligible position from last maximum */
  int numMax;                   /* number of maxima found so far */
  long seqIndex;                /* index into array of motif probabilities for current sequence */   


  if (spaceNum!=1 && spaceNum!=2){
    Rprintf("sortIntMotProbs: spaceNum must be 1 or 2. Got %d. Exiting ...\n", spaceNum);
    return(0);
  }

  numMax        = 0;
  numSeqs       = dataSet->numSeqs;
  maxSeqLength  = dataSet->maxSeqLength;
  intMotProbs   = (spaceNum==1) ? dataSet->intMotProbs1 : dataSet->intMotProbs2;
  maxMotProbs   = dataSet->maxMotProbs;
  motProbLength = (1 + dataSet->revComp) * maxSeqLength;
  

  /* find maximum motif probabilities on all sequences such that no maxima are within W nucleotides of each other */
  if (global){

    for (seqNum=0; seqNum<numSeqs; seqNum++){
 
      seqLength = dataSet->samples[seqNum].length;
      seq       = dataSet->samples[seqNum].seq;
      lastPos   = seqLength - width;
      seqIndex  = seqNum*motProbLength;
      maxProb   = intMotProbs[seqIndex].prob;

      if (seqLength < width) continue;
      
      maxMotProbs[numMax].seq  = seq;
      maxMotProbs[numMax].prob = maxProb;
      nextPos                  = MIN(width,lastPos+1);
      
      for (pos=0; pos<=lastPos; pos++){
	prob = intMotProbs[seqIndex+pos].prob;
	
	if (pos==nextPos) numMax++;
	
	if (pos==nextPos || prob > maxProb){
	  maxProb = prob;
	  maxMotProbs[numMax].seq  = seq + pos;
	  maxMotProbs[numMax].prob = prob;
	  nextPos = MIN(pos+width,lastPos+1);

	  
	}
      }
      numMax++;


      /* check reverse complement if necessary */
      if (dataSet->revComp){
	seqIndex     = seqNum*motProbLength + maxSeqLength;
	seq       = dataSet->samples[seqNum].rcSeq;
	maxProb   = intMotProbs[seqIndex].prob;

	maxMotProbs[numMax].seq  = seq;
	maxMotProbs[numMax].prob = maxProb;
	nextPos                  = MIN(width,lastPos+1);
	

	for (pos=0; pos<=lastPos; pos++){
	  prob = intMotProbs[seqIndex+pos].prob;

	  if (pos==nextPos) numMax++;
	  
	  if (pos==nextPos || prob > maxProb){
	    maxProb = prob;
	    maxMotProbs[numMax].seq  = seq + pos;
	    maxMotProbs[numMax].prob = prob;
	    nextPos = MIN(pos+width, lastPos+1);
	  }
	}
	numMax++;
      }

    }	
  }else{

    /* find maximum motif probabilities for each sequence separately */
    for (seqNum=0; seqNum<numSeqs; seqNum++){

      seqLength = dataSet->samples[seqNum].length;
      seq = dataSet->samples[seqNum].seq;
      seqIndex     = seqNum*motProbLength;

      if (seqLength < width) continue;

      maxMotProbs[seqNum].prob = intMotProbs[seqIndex].prob;
      maxMotProbs[seqNum].seq  = seq;


      for (pos=0; pos<seqLength - width + 1; pos++){

	if (intMotProbs[seqIndex + pos].prob > maxMotProbs[seqNum].prob){
	  maxMotProbs[seqNum].prob        = intMotProbs[seqIndex + pos].prob;
	  maxMotProbs[seqNum].seq         = seq + pos;
	}
      }	

      if (dataSet->revComp){
	seq    = dataSet->samples[seqNum].rcSeq;
	seqIndex  = seqNum*motProbLength + maxSeqLength;

	for (pos=0; pos<seqLength - width + 1; pos++){
	  
	  if (intMotProbs[seqIndex + pos].prob > maxMotProbs[seqNum].prob){
	    maxMotProbs[seqNum].prob        = intMotProbs[seqIndex + pos].prob;
	    maxMotProbs[seqNum].seq         = seq + pos;
	  }
	}	

      }

      numMax++;
    }

  }

  /* finally sort the identified maxima */
  qsort((void*) maxMotProbs,numMax,sizeof(INTMOTPROB),compIntMotProbs);

  if (DEBUG>=4) Rprintf("sortIntComp: numMax = %d and global = %d\n", numMax, global);

  return numMax;

}


/**********************************************************************/
/*
  insertSv

  Insert starting value with position weight matrix 'pwm' and score 
  'score' in position 'where' of array 'svTheta'.

 
*/
/**********************************************************************/
int insertSv(
	      THETA *svTheta,    /* array of starting values */
	      int where,         /* where in this array are we inserting a starting value? */
	      double *pwm,       /* position weight matrix of starting value to be inserted */
	      double score,      /* score of this starting value */
	      int width          /* width of the position weight matrix */
	      )
{

  int pos;   /* index for looping over positions in PWM */
  int nuc;   /* index for looping over nucleotides*/


  svTheta[where].width      = width;
  svTheta[where].score      = score;

  /* copy PWM */
  for (pos=0; pos<width; pos++){
    for (nuc=0; nuc<4; nuc++){
      svTheta[where].pwm[PWM(pos,nuc)] = pwm[PWM(pos,nuc)];
    }
  }

  if (DEBUG>=3){
    Rprintf("insertSv: inserting starting value with log-LR = %lf \n", score);
    PrintDoubleMatrix2File(stderr,pwm,4,width);
  }

  return(1);
}

/**********************************************************************/
/*
  compTheta

  Compare to staring values on the basis of their 'score'.

  Returns -1 if first score is greater
           1 if first score is less
	   0 otherwise
 
*/
/**********************************************************************/
int compTheta(
	      const void *x,   /* first starting value */ 
	      const void *y    /* second starting value */
	      )
{

  THETA *t1, *t2;

  t1 = (THETA *) x;
  t2 = (THETA *) y;

  if (t1->score > t2->score) return -1;
  if (t2->score > t1->score) return 1;
  return 0;
}


/**********************************************************************/
/*
  checkStart

  After having calculated the motif probabilities for an initial starting
  value, form new starting values by aligning different numbers of sites
  based on their motif probabilities.
  Keep the one with the smallest E-value and check if it is better than
  the worst starting value recorded so far.
  If so, starting value is inserted into the array 'svTheta' of the
  appropriate starting value space.
  
 
*/
/**********************************************************************/
int checkStart(
		DATASET *dataSet,   /* the dataset */
		int spaceNum,       /* are motif probs stored in intMotProb1 or intMotProb2? */
		int width           /* width of motif */
		)
{

  int modTypeNum;          /* index of current model type */  
  int numModTypes;         /* number of model types */
  int nnSites;             /* number of canidates values for expected number of motif occurrences */          
  int nSitesNum;           /* index of current candidate value */
  int siteNum;             /* index for looping over sites to be aligned */
  int numSites;            /* current number of aligned sites */
  MTYPE mType;             /* current model type */
  int conSetNum;           /* number of current constraint set */
  int numConSets;          /* number of constraint sets */
  BOOLEAN global;          /* sort motif probabilities globally or for each sequence separately? */
  THETA *svTheta;          /* starting value */
  SVSPACE *svSpace;        /* starting value space for a given model type */
  int *motifCounts;        /* nucleotide counts for the current alignment */
  double *pwm;             /* position weight matrix created from alignment */
  int *seq;                /* pointer to current sequence */
  double *backFreq;        /* nucleotide frequencies in background */
  int numSv;               /* number of starting values to be found */
  int pos;                 /* current position in motif */
  int nuc;                 /* current nucleotide at that position */
  double f;                /* observed nucleotide frequency in alignment */
  double p;                /* nucleotide frequency in background */
  double llr;              /* log-likelihood ratio for hypothesis test that alignment was generated by background distribution */
  double intensity;        /* value of intensity parameter */
  MODEL *mod;              /* interpreted constraint set */
  int numMax;              /* number of maxima found by sorting function */
  double log_pop;          /* log-transformed product of column p-values */
  double log_pv;           /* log-transformed p-value for column-specific hypothesis test */
  int jj;                  /* index */ 
  

  numConSets     = dataSet->numConSets;
  numModTypes    = dataSet->numModTypes;
  motifCounts    = dataSet->motifCounts;
  backFreq       = dataSet->nucFreq;
  pwm            = dataSet->pwm1;

  if (spaceNum!=1 && spaceNum!=2){
    Rprintf("checkStart: spaceNum must be 1 or 2. Got %d. Exiting ...\n", spaceNum);
    return(0);
  }

  /* loop over candidate model types */
  for (modTypeNum=0; modTypeNum<numModTypes; modTypeNum++){
    svSpace   = dataSet->svSpace + modTypeNum;
    nnSites   = svSpace->nnSites;
    mType     = svSpace->mType;
    numSv     = svSpace->numSv;
    
    global    = mType==TCM;
    
    /* obtain sorted maxima of motif probabilities */
    numMax    = sortIntMotProbs(dataSet,spaceNum,global, width);
    
    
    /* initialize motif counts to zero */
    for (pos=0;pos<width;pos++){
      for (nuc=0;nuc<4;nuc++){
	    motifCounts[PWM(pos,nuc)] = 0;
      }
    }
     
    /* align top sites one at a time */
    for (nSitesNum=0, siteNum=0; nSitesNum<nnSites; nSitesNum++){
      numSites = svSpace->numSites[nSitesNum];


      /* don't score this starting point if we didn't find enough maxima */
      if (numMax < numSites) continue;

      intensity = getIntensity(mType, width, numSites, dataSet);

      /* add new sites to old alignment */
      for (;siteNum<numSites; siteNum++){
	seq = dataSet->maxMotProbs[siteNum].seq;
	for (pos=0;pos<width;pos++){
	  if (seq[pos]<4){
	    /* regular character */
	    motifCounts[PWM(pos,seq[pos])]++;
	  }else{
	    /* add background frequencies for unknown 'X' */
	    for (jj=0; jj<4; jj++) motifCounts[PWM(pos,jj)] += dataSet->nucFreq[jj];
	  }
	}
      }
      
      if (DEBUG>=3){
	PrintIntMatrix2File(stderr,motifCounts,4,width);
	Rprintf("\n");
      }


      log_pop = 0;	
      for (pos=0;pos<width;pos++){

	/* compute log likelihood for position pos */
	llr = 0;
	for (nuc=0;nuc<4;nuc++){
	  f = ((double) motifCounts[PWM(pos,nuc)])/((double) numSites);
	  p = backFreq[nuc];
	  llr += (f&&p) ? f *(log(f) - log(p)) : 0; 
	}
	llr *= numSites;	         	        /* convert entropy to ll */ 
	RND(llr, RNDDIG, llr);       	       	        /* round to RNDDIG places */
	log_pv = get_llr_pv(llr, numSites, 1, LLR_RANGE, 1.0, 4, backFreq); 
	if (DEBUG>=4) Rprintf("column p-value:%lf\n",log_pv);
	log_pop += log_pv;	
	
      }

      /* don't need to calculate p-value of this product of uniforms since width is fixed */

      /* create position weight matrix from alignment */
      for (pos=0; pos<width; pos++){
	for (nuc=0; nuc<4;nuc++){
	  /* log-likelihood function is not differentiable for p=0 */
	  pwm[PWM(pos,nuc)] = MAX(1e-16,(double) motifCounts[PWM(pos,nuc)]/((double)numSites));
	}
      }
      
      if (DEBUG>=3) PrintDoubleMatrix2File(stderr,pwm,4,width);
      
      /* loop over candidate constraint sets */
      for (conSetNum=0; conSetNum<numConSets; conSetNum++){
      
	/* set model for chosen width, modelType, and constraint set */
	mod = dataSet->mod + MOD(width, modTypeNum, conSetNum);

	/* check if PWM satisfies constraints */
	if (!satisfiesConstraints(pwm, width, mod)) continue;
	
	svTheta = svSpace->svTheta + SV(width,conSetNum,nSitesNum,0);
	if (DEBUG>=3 && nSitesNum==nnSites-1) Rprintf("score = %lf vs. score3 = %lf\n", -log_pop, svTheta[numSv-1].score);
	
	
	/* record starting value if is unique and has a better score than currently worst recorded one */
	if (-log_pop > svTheta[numSv-1].score){
	  
	  if (!isDup(pwm, width, svTheta, numSv)){
	    if (!insertSv(svTheta, numSv-1, pwm, -log_pop, width)) return(0);
	    qsort((void*) svTheta,numSv,sizeof(THETA), compTheta); 	    

	    if (DEBUG>=3){
	      Rprintf("New starting value for width %d and numsites = %d with score = %lf\n", width, numSites, -log_pop);
	      PrintDoubleMatrix2File(stderr, pwm, 4, width);
	      Rprintf("\n");
	    }
	  }
	}
	
      }
    }
  }

  return(1);
}


/**********************************************************************/
/*
  isDup
  
  Check if position weight matrix 'pwm' is already among array of
  recorded starting values.

  Return TRUE if this is the case, FALSE otherwise.
  
 
*/
/**********************************************************************/
int isDup(
	  double *pwm,        /* candidate PWM */
	  int width,          /* its width */
	  THETA *svTheta,     /* array of recorded starting values */
	  int numSv           /* number of recorded starting values */
	  )
{
	
  int startNum;   /* index of current recorded starting value */
  double *svPwm;  /* pwm of this starting value */
  double epsilon; /* declare duplicate if Euclidean distance between two PWM is less than epsilon */
  BOOLEAN dup;    /* is this a duplicate? */

  epsilon = 1e-4;

  dup = FALSE;
  for (startNum=0;startNum<numSv && dup==0;startNum++){
    svPwm = svTheta[startNum].pwm;
    dup = euclDist(svPwm, pwm, width) < epsilon;
  }

  return dup;
}
 

/**********************************************************************/
/*
  getBestNSites
  
  For a given position weight matrix, find the expected number of sites
  that minimizes the E-value of the corresponding alignment of sites
  with the highest motif probabilities.
  
 
*/
/**********************************************************************/
int getBestNSites(
		   THETA *theta,         /* parameter estimates */
		   DATASET *dataSet      /* the dataset */
		   )
{
  int nnSites;             /* number of such candidate values */
  MTYPE mType;             /* current model type */
  int numSeqs;             /* number of sequences */
  int numConSets;          /* number of constraint sets */
  BOOLEAN global;          /* sort motif probabilities globally or for each sequences separately? */
  int numModTypes;         /* number of model types */
  int siteNum;             /* index of current site to be added to the alignment */
  int numSites;            /* number of sites in the alignment */
  int *motifCounts;        /* nucleotide counts for current alignment */
  int *seq;                /* current sequence */
  double *backFreq;        /* nucleotide frequencies in background distribution */
  int pos;                 /* position in motif */
  int nuc;                 /* nucleotide */
  double f;                /* observed nucleotide frequency in alignment */
  double p;                /* nucleotide frequency in background */
  double llr;              /* log-likelihood ratio for hypothesis test that alignment was generated by background distribution */
  double log_pop;          /* log-transformed product of column p-values */
  double log_pv;           /* log-transformed p-value for column-specific hypothesis test */
  double log_nalign;       /* log number of possible alignments */
  double logEval;          /* log E-value */
  double bestLogEval;      /* best log E-value so far */
  int bestNumSites;        /* best number of sites to align so far */
  double *bestPWM;         /* best PWM so far */
  int width;               /* width of motif */
  int seqNum;              /* current sequence index */
  int start;               /* current position in sequence */
  int maxNumSites;         /* maximum number of sites to align */
  int seqLength;           /* length of current sequence */
  int jj;                  /* index */

  numConSets     = dataSet->numConSets;
  numSeqs        = dataSet->numSeqs;
  numModTypes    = dataSet->numModTypes;
  motifCounts    = dataSet->motifCounts;
  backFreq       = dataSet->nucFreq;
  bestPWM        = dataSet->pwm1;
  width          = theta->width;


  /* calculate posterior probabilities */
  if (!getPostProbs(theta,dataSet)) return(0);

  /* find appropriate maximum number of sites to align for current model type */
  mType   = theta->mType;
  switch (mType){
  case OOPS:
    /* do nothing for OOPS model */
    return(1);
  case ZOOPS:
    maxNumSites = MIN(dataSet->maxNumSites,dataSet->numSeqs);
    break;
  case TCM:
  default:
    maxNumSites = dataSet->maxNumSites;
    break;
  }
  nnSites = maxNumSites - dataSet->minNumSites + 1;
  

  /* sort posterior probabilities */
  global  = mType==TCM;
  if (!sortPostProbs(dataSet, global)) return(0);
        
  /* initialize nucleotide counts to zero */
  for (pos=0;pos<width;pos++){
    for (nuc=0;nuc<4;nuc++){
      motifCounts[PWM(pos,nuc)] = 0;
    }
  }

  bestNumSites = dataSet->minNumSites;
  bestLogEval   = BIG;
  for (numSites=dataSet->minNumSites, siteNum=0; numSites <= maxNumSites; numSites++){

    if (DEBUG>=3) Rprintf("siteNum = %d and numSites = %d\n", siteNum, numSites);

    /* add new sites to the alignment */
    while (siteNum<numSites){

      seqNum    = dataSet->maxPostProbs[siteNum].seqNum;
      seqLength = dataSet->samples[seqNum].length;
      start     = dataSet->maxPostProbs[siteNum].start;

      seq       = dataSet->maxPostProbs[siteNum].revComp ? dataSet->samples[seqNum].rcSeq + seqLength - start -width : dataSet->samples[seqNum].seq + start;

      if (DEBUG>=3){
	Rprintf("Site %d with prob %lf (revComp = %d):\n", siteNum, dataSet->maxPostProbs[siteNum].prob,
                                                                  dataSet->maxPostProbs[siteNum].revComp);
	PrintIntMatrix2File(stderr,seq, 1, width);
	Rprintf("\n");
      }

      /* add site to the alignment */
      for (pos=0;pos<width;pos++){ 
	if (seq[pos]<4){
	  motifCounts[PWM(pos,seq[pos])]++;
	}else{
	  for (jj=0; jj<4; jj++) motifCounts[PWM(pos,jj)] += dataSet->nucFreq[jj];
	}
      }
      siteNum++;
      if (DEBUG>=3) Rprintf("at end of loop siteNum = %d\n", siteNum);
    }

    if (DEBUG>=3){
      Rprintf("\nnumSites = %d:\n", numSites);
      PrintIntMatrix2File(stderr, motifCounts,4,width);
    }


    /* compute log likelihood ratio */
    log_pop = 0;			
    for (pos=0;pos<width;pos++){
      
      /* compute log likelihood for position pos */
      llr = 0;
      for (nuc=0;nuc<4;nuc++){
	f = ((double) motifCounts[PWM(pos,nuc)])/((double) numSites);
	p = backFreq[nuc];
	llr += (f&&p) ? f *(log(f) - log(p)) : 0; 
      }
      /* convert entropy to ll */ 
      llr *= numSites;
      /* round to RNDDIG places */
      RND(llr, RNDDIG, llr);      
      log_pv = get_llr_pv(llr, numSites, 1, LLR_RANGE, 1.0, 4, backFreq); 
      log_pop += log_pv;	

    }
   
    /* added this 10/14/05 */
    /* get p-value of product of p-values */
    log_pv = log_qfast(width, log_pop);	
 
    /* (log) number of possible alignments of this width and model type */
    log_nalign = get_log_nalign(mType, theta->width, numSites, dataSet);

    if (DEBUG>=3) Rprintf("log_pv = %lf and log_nalign = %lf\n", log_pv, log_nalign);

    logEval = log_pv + log_nalign;

    RND(logEval, RNDDIG, logEval);

    if (DEBUG>=3) Rprintf("numSites = %d logEval = %lf\n", numSites, logEval);

    /* new best E-value? */
    if (RNDEPS < bestLogEval - logEval){
      bestNumSites  = numSites;
      bestLogEval   = logEval;
          
      /* copy position weight matrix */
      for (pos=0; pos<width; pos++){
	for (nuc=0; nuc<4;nuc++){
	  bestPWM[PWM(pos,nuc)] = (double) motifCounts[PWM(pos,nuc)]/((double)numSites);
	}
      }
      
      if (DEBUG>=3){
	Rprintf("New bestNumSites = %d, bestLogEval = %lf:\n", bestNumSites, bestLogEval);
	if (DEBUG>=3) PrintDoubleMatrix2File(stderr,bestPWM,4,width);
      }
    }

  }

  /* update number of sites, intensity, and E-value of parameter estimates */
  theta->numSites  = bestNumSites;
  theta->intensity = getIntensity(theta->mType, theta->width, theta->numSites, dataSet);
  theta->logEval   = bestLogEval;



  if (DEBUG>=3){
    Rprintf("bestNumSites = %d and bestLogEval = %lf:\n", theta->numSites, theta->logEval);
    PrintDoubleMatrix2File(stderr,theta->pwm, 4, theta->width);
  }

  return(1);
}



/**********************************************************************/
/*
  getIntensity
  
  For a given expected number of motif occurrences and  model type, 
  find the corresponding intensity parameter.

  Returns that intensity parameter.
  
 
*/
/**********************************************************************/
double getIntensity(
		    MTYPE mType,          /* model type */
 		    int width,            /* width of motif */
		    int numSites,         /* expected number of motif occurrences */
		    DATASET *dataSet      /* the dataset */
		    )
{


  double intensity;      /* intensity parameter */
  int minWidth;          /* minimum motif width to consider */
  

  minWidth = dataSet->minWidth;
  switch(mType){
  case OOPS:
    intensity = 1;
    break;

  case ZOOPS:
    intensity = (double) numSites/(double) dataSet->numSeqs;
    break;
    
  case TCM:
    intensity = (double) numSites / (double) dataSet->totalNucs;
    break;

  case CUT:
  case OVER:
    intensity = 1 - pow(1-1/((double)dataSet->numCuts[width-minWidth]), numSites);
    break;
  default:
    Rprintf("getIntensity: mType %d not recognized. Exiting ...\n", mType);
    return(0);
    break;
  }

  return intensity;
}


/**********************************************************************/
/*
  completeSv
  
  Add additional information to starting value beyond just PWM and 
  score.
  This is done for all starting values.
  
 
*/
/**********************************************************************/
int completeSv(
		DATASET *dataSet   /* the dataset */
		)
{

  int conSetNum;       /* index of current constraint set */
  int numConSets;      /* number of candidate constraint sets */
  MODEL *mod;          /* interpreted constraint set */
  int ppNum;           /* index for looping over IC profile parameters */
  int nSitesNum;       /* index of current expected number of motif occurrences */
  int nnSites;         /* number of candidate values for this */
  int numSv;           /* number of starting values */
  int startNum;        /* index of current starting value */
  THETA *svTheta;      /* a starting value */
  SVSPACE *svSpace;    /* starting value space for given model type */

  int width;           /* current motif width */
  int minWidth;        /* minimum motif width */
  int maxWidth;        /* maximum motif width */
  int numModTypes;     /* number of model types */
  int modTypeNum;      /* index of current model type */
  MTYPE mType;         /* model type */

  double intensity;    /* intensity parameter */
  int siteNum;         /* current site number */
  int numSites;        /* number of sites */

  numConSets     = dataSet->numConSets;
  minWidth       = dataSet->minWidth;
  maxWidth       = dataSet->maxWidth;
  numModTypes    = dataSet->numModTypes;

  /* loop over all model types */
  for (modTypeNum=0; modTypeNum<numModTypes; modTypeNum++){
    svSpace     = dataSet->svSpace + modTypeNum;
    mType       = svSpace->mType;
    nnSites     = svSpace->nnSites;
    numSv       = svSpace->numSv;


    /* loop over all widths */
    for (width=minWidth; width<=maxWidth; width++){


      for (nSitesNum=0, siteNum=0; nSitesNum<nnSites; nSitesNum++){
	numSites = svSpace->numSites[nSitesNum];

	/* get intensity parameter */
	intensity = getIntensity(mType, width, numSites, dataSet);
	
	/* loop over constraint sets */
	for (conSetNum=0; conSetNum<numConSets; conSetNum++){

	  /* set model for chosen width, modelType, and constraint set */
	  mod = dataSet->mod + MOD(width, modTypeNum, conSetNum);

	  /* loop over starting value */
	  for (startNum=0;startNum<numSv;startNum++){
      
	    /* get pointer to this starting value */
	    svTheta = svSpace->svTheta + SV(width, conSetNum, nSitesNum, startNum);;

	    svTheta->numSites = numSites;
	    svTheta->intensity = intensity;
	    svTheta->width = width;
	    svTheta->mType = mType;
	    svTheta->conMod = mod;

	    /* create starting values for profile parameters */
	    svTheta->numProfParms = mod->numProfParms;
	    for (ppNum=0; ppNum<mod->numProfParms; ppNum++){
	      svTheta->profParms[ppNum] = 
		0.5*(mod->low[mod->profParmIndex + ppNum] + mod->up[mod->profParmIndex + ppNum]);
	    }      
	    
	  }
	}
      }
    }
  }

  return(1);
}


/**********************************************************************/
/*
  switchMod
  
  Switch model type of theta to mType and update intensity parameter.
 
*/
/**********************************************************************/
int switchMod(
	       THETA *theta,           /* parameter estimates */
	       MTYPE mType,            /* model type to switch to */ 
	       DATASET *dataSet        /* the dataset */
	       )
{

  theta->mType = mType;
  theta->intensity = getIntensity(theta->mType,theta->width,theta->numSites,dataSet);

  return(1);   
}
