/*****************************************************************************
*  			     					             *
*      cosmo - Constrained search for motifs in DNA sequences                *
*      Copyright (C) 2006  Oliver Bembom (bembom@berkeley.edu)               *
*                                                                            *
*      This file is part of cosmo.                                           *
*                                                                            *
*      cosmo is free software; you can redistribute it and/or                *
*      modify it under the terms of the GNU Lesser General Public            *
*      License as published by the Free Software Foundation; either          *
*      version 2.1 of the License, or (at your option) any later version.    *
*                                                                            *
*      cosmo is distributed in the hope that it will be useful,              *
*      but WITHOUT ANY WARRANTY; without even the implied warranty of        *
*      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU     *
*      Lesser General Public License for more details.                       *
*                                                                            *
*      You should have received a copy of the GNU Lesser General Public      *
*      License along with this program; if not, write to the Free Software   *
*      Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,            *
*      MA  02110-1301, USA.                                                  *
*                                                                            *
*                                                                            *
*****************************************************************************/


/*****************************************************************************
*                                                                            *
*      This module contains functions for freeing memory for various         * 
*      structures                                                            *
*                                                                            *
*****************************************************************************/


#include <R_ext/PrtUtil.h>
#include <R_ext/Memory.h>
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <float.h>
#include "macros.h"
#include "cosmo.h"
#include "free.h" 
  

/**********************************************************************/
/*
  dataSetFree

  Free memory allocated for this dataset.
	
*/
/**********************************************************************/
void dataSetFree(
		 DATASET *dataSet     /* the dataset */
		 )
{

  if (dataSet==NULL) return;

  /* these objects can be freed the normal way */
  myfree(dataSet->nucFreq);
  myfree(dataSet->mTypes);
  myfree(dataSet->modTypeName);
  myfree(dataSet->backProb);
  myfree(dataSet->logLiks);
  myfree(dataSet->excludeCv);
  myfree(dataSet->excludeCv2);  
  myfree(dataSet->refThetaConv);
  myfree(dataSet->pwm1);
  myfree(dataSet->pwm2);
  myfree(dataSet->lPwm1);
  myfree(dataSet->lPwm2);
  myfree(dataSet->cumProb);
  myfree(dataSet->motifCounts);  
  myfree(dataSet->intMotProbs1);
  myfree(dataSet->intMotProbs2);
  myfree(dataSet->maxMotProbs);
  myfree(dataSet->postProbs);
  myfree(dataSet->maxPostProbs); 
  myfree(dataSet->transMats);
  myfree(dataSet->allCrits);
  myfree(dataSet->intNumBasePairs);
  myfree(dataSet->intFirst);
  myfree(dataSet->intLast);
  myfree(dataSet->line);
  myfree(dataSet->word);
  myfree(dataSet->first);


  /* these structure need to memory free inside of them */
  svSpaceFree(dataSet->svSpace, dataSet->numModTypes); 
  sampleFree(dataSet->samples, dataSet->numSeqs);
  sampleFree(dataSet->backSamples, dataSet->numBackSeqs);
  conSetFree(dataSet->cSets, dataSet->numConSets);
  modFree(dataSet->mod, dataSet->numModels);
  thetaFree(dataSet->refThetas, dataSet->nRefThetas);
  thetaFree(dataSet->svTheta,1);
  thetaFree(dataSet->theta,1);
  thetaFree(dataSet->theta1,1);
  thetaFree(dataSet->theta2,1);


}

/**********************************************************************/
/*
  svSpaceFree

  Free memory allocated inside n SVSPACE structs

*/
/**********************************************************************/
void svSpaceFree(
		 SVSPACE *svSpace, /* pointer to the first SVSPACE struct */
		 long n            /* number of structs to free memory for */
		 )
{
  int num;              /* index for iterating over structs */

  if (svSpace==NULL) return;

  for(num=0; num<n;num++){

    myfree(svSpace[num].svLogLR);
    myfree(svSpace[num].numSites);
    thetaFree(svSpace[num].svTheta, svSpace[num].totSv);  
  }

  myfree(svSpace);
}

/**********************************************************************/
/*
  thetaFree

  Free memeory allocated inside n THETA structs

*/
/**********************************************************************/
void  thetaFree(
		THETA *theta,      /* pointer to first THETA struct */
		long n             /* number of structs to free memory for */
		)
{
  int num;           /* index for iterating over structs */

  if (theta==NULL) return;

  for (num=0; num<n; num++){

    myfree(theta[num].pwm);
    myfree(theta[num].profParms);
    myfree(theta[num].posKL);
    myfree(theta[num].posIC);
    myfree(theta[num].colScores);

  }

  myfree(theta);
}

/**********************************************************************/
/*
  sampleFree

  Free memory inside n SAMPLE structs.

*/
/**********************************************************************/
void sampleFree(
		SAMPLE *sample,  /* pointer to first sample */
		long n           /* number of structs to free memory for */
		)
{

  int num;               /* index for iterating over structs */

  if (sample==NULL) return;

  for (num=0; num<n; num++){

    myfree(sample[num].name);
    myfree(sample[num].seq);
    myfree(sample[num].rcSeq);
    myfree(sample[num].numCuts);
    myfree(sample[num].logCumBack);
    myfree(sample[num].backLR);;
    myfree(sample[num].cumProbBack);
    myfree(sample[num].mProbs1);
    myfree(sample[num].rcmProbs1);
    myfree(sample[num].mProbs2);
    myfree(sample[num].rcmProbs2);
    myfree(sample[num].seqProbs);
    myfree(sample[num].rcSeqProbs);
    myfree(sample[num].cutLRs); 

  }

  myfree(sample);
}

/**********************************************************************/
/*
  conSetFree

  Free memory inside n CONSET structs.

*/
/**********************************************************************/
void conSetFree(
		CONSET *conSets,   /* pointer to first CONSET struct */
		long n            /* number of structs to allocate memory for */
		)
{
    
  int cSetNum;             /* index for iterating over structs */

  if (conSets==NULL) return;

  for (cSetNum=0; cSetNum<n;cSetNum++){
    
    intConFree(conSets[cSetNum].intCons, conSets[cSetNum].numIntFreqCons);
    linConFree(conSets[cSetNum].linCons);
  }

  myfree(conSets);

}

/**********************************************************************/
/*
  posFreqConFree

  Free memory for linked POSFREQCON structs.

*/
/**********************************************************************/
void posFreqConFree(
		    POSFREQCON * posFreqCon       /* pointer to first struct */
		    )
{

  if (posFreqCon==NULL) return;

  if (posFreqCon->next) posFreqConFree(posFreqCon->next);
  myfree(posFreqCon);

}

/**********************************************************************/
/*
  intFreqConFree

  Free memory for linked INTFREQCON structs.

*/
/**********************************************************************/
void intFreqConFree(
		    INTFREQCON * intFreqCon       /* pointer to first struct */
		    )
{

  if (intFreqCon==NULL) return;

  if (intFreqCon->next) intFreqConFree(intFreqCon->next);
  myfree(intFreqCon);

}


/**********************************************************************/
/*
  linConFree

  Free memory for linked LINCONCON structs.

*/
/**********************************************************************/
void linConFree(
		LINCON * linCon       /* pointer to first struct */
		)
{
  
  if (linCon==NULL) return;

  if (linCon->next) linConFree(linCon->next);
  myfree(linCon);

}


/**********************************************************************/
/*
  intConFree

  Free memory inside n INTCON structs.

*/
/**********************************************************************/
void intConFree(
		INTCON *intCons,  /* pointer to first struct */
		long n            /* number of structs to allocate memory for */
		)
{
    
  int intConNum;        /* index for iterating over structs */

  if (intCons==NULL) return;

  for (intConNum=0;intConNum<n;intConNum++){

    intFreqConFree(intCons[intConNum].intFreqCons);
    posFreqConFree(intCons[intConNum].posFreqCons);
  }

  myfree(intCons);
}


/**********************************************************************/
/*
  modFree

  Free memory inside n MODEL structs.

*/
/**********************************************************************/
void modFree(
	     MODEL *mod,   /* pointer to first struct */
	     long n        /* number of structs to allocate memory for */
	     )
{
    
  int modNum;          /* index for iterating over structs */

  if (mod==NULL) return;

  for (modNum = 0; modNum < n; modNum++){

    myfree(mod[modNum].linConMatrix);
    myfree(mod[modNum].low);
    myfree(mod[modNum].up);
    myfree(mod[modNum].nlCons);
  }
    
  myfree(mod);

}



