/*****************************************************************************
*  			     					             *
*      cosmo - Constrained search for motifs in DNA sequences                *
*      Copyright (C) 2006  Oliver Bembom (bembom@berkeley.edu)               *
*                                                                            *
*      This file is part of cosmo.                                           *
*                                                                            *
*      cosmo is free software; you can redistribute it and/or                *
*      modify it under the terms of the GNU Lesser General Public            *
*      License as published by the Free Software Foundation; either          *
*      version 2.1 of the License, or (at your option) any later version.    *
*                                                                            *
*      cosmo is distributed in the hope that it will be useful,              *
*      but WITHOUT ANY WARRANTY; without even the implied warranty of        *
*      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU     *
*      Lesser General Public License for more details.                       *
*                                                                            *
*      You should have received a copy of the GNU Lesser General Public      *
*      License along with this program; if not, write to the Free Software   *
*      Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,            *
*      MA  02110-1301, USA.                                                  *
*                                                                            *
*                                                                            *
*****************************************************************************/


/*****************************************************************************
*                                                                            *
*      This module contains various functions for allocating memory to       *
*      arrays and printing the content of an array to the screen or to a     *
*      file                                                                  *     
*                                                                            *
*****************************************************************************/

#include <R_ext/PrtUtil.h>
#include <R_ext/Memory.h>
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <float.h>
#include "matLib.h"
  
/**********************************************************************/
/*
  charMalloc

  Allocate memory for n characters.

*/
/**********************************************************************/
char  *CharMalloc(
		  long n    /* number of characters for which to allocate memory */
		  )
{
  char *x;

  if (n <= 0){
    Rprintf("CharMalloc: Don't allocate n = %ld objects!\n", n);
    exit(1);
  }

  if ((x=(char *)R_alloc((unsigned)(n),(sizeof(char))))==NULL){
      Rprintf("ERROR:Couldn't get %ld characters \n",n);
      exit(1);
  }

  return(x);
}

/**********************************************************************/
/*
  DoubleMalloc

  Allocate memory for n doubles.

*/
/**********************************************************************/
double *DoubleMalloc(
		     long n     /* number of doubles for which to allocate memory */
		     )
{
  double *x;

  if (n <= 0){
    Rprintf("DoubleMalloc: Don't allocate n = %ld objects!\n", n);
    exit(1);
  }

  if ((x=(double *)R_alloc((unsigned)(n),(sizeof(double))))==NULL){
      Rprintf("ERROR:Couldn't get %ld doubles \n",n);
      exit(1);
  }

  return(x);
}

/**********************************************************************/
/*
  p2DoubleMalloc

  Allocate memory for n pointers to double.

*/
/**********************************************************************/
double **p2DoubleMalloc(
			long n    /* number of pointers to double for which to allocate memory */
			)
{
  double **x;

  if (n <= 0){
    Rprintf("p2DoubleMalloc: Don't allocate n = %ld objects!\n", n);
    exit(1);
  }

  if ((x=(double **)R_alloc((unsigned)(n),(sizeof(double *))))==NULL){
      Rprintf("ERROR:Couldn't get %ld pointers to double \n",n);
      exit(1);
  }

  return(x);
}

/**********************************************************************/
/*
  LongMalloc

  Allocate memory for n longs.

*/
/**********************************************************************/
long *LongMalloc(
		 long n    /* number of longs for which to allocate memory */
		 )
{
  long *x;

  if (n <= 0){
    Rprintf("LongMalloc: Don't allocate n = %ld objects!\n", n);
    exit(1);
  }

  if ((x=(long *)R_alloc((unsigned)(n),(sizeof(long))))==NULL){
      Rprintf("ERROR:Couldn't get %ld longs \n",n);
      exit(1);
  }

  return(x);
}

/**********************************************************************/
/*
  IntMalloc

  Allocate memory for n ints.

*/
/**********************************************************************/
int *IntMalloc(
	       long n     /* number of ints for which to allocate memory */
	       )
{
  int *x;

  if (n <= 0){
    Rprintf("IntMalloc: Don't allocate n = %ld objects!\n", n);
    exit(1);
  }

  if ((x=(int *)R_alloc((unsigned)(n),(sizeof(int))))==NULL){
      Rprintf("ERROR:Couldn't get %ld ints \n",n);
      exit(1);
  }

  return(x);
}

/**********************************************************************/
/*
  pIntMalloc

  Allocate memory for n pointers to int.

*/
/**********************************************************************/
int **pIntMalloc(
		 long n      /* number of pointers to int for which to allocate memory */
		 )
{
  int **x;

  if (n <= 0){
    Rprintf("pIntMallocc: Don't allocate n = %ld objects!\n", n);
    exit(1);
  }

  if ((x=(int **)R_alloc((unsigned)(n),(sizeof(int *))))==NULL){
      Rprintf("ERROR:Couldn't get %ld pointers to int \n",n);
      exit(1);
  }

  return(x);
}

 
/**********************************************************************/
/*
  PrintDoubleMatrix

  Print a nrow by ncol matrix of doubles to the screen.

*/
/**********************************************************************/
void PrintDoubleMatrix(
		       double *x,    /* pointer to the matrix of doubles */
		       long nrow,    /* number of rows */
		       long ncol     /* number of columns */
		       )
{
  long i,j;

  for (i=0;i<nrow;i++){
    for(j=0;j<ncol;j++){
      Rprintf("%.4lf ",x[j*nrow+i]); 
    }
    Rprintf("\n");
  }
}

/**********************************************************************/
/*

  PrintDoubleMatrixE

  Print a nrow by ncol matrix of doubles in scientif format to the 
  screen.

*/
/**********************************************************************/
void PrintDoubleMatrixE(
			double *x,    /* pointer to the matrix of doubles */
			long nrow,    /* number of rows */
			long ncol     /* number of columns */
			)
{
  long i,j;
  for (i=0;i<nrow;i++){
    for(j=0;j<ncol;j++){
      Rprintf("%.4e ",x[j*nrow+i]); 
    }
    Rprintf("\n");
  }
}

/**********************************************************************/
/*

  PrintLongMatrix

  Print a nrow by ncol matrix of longs to the screen.

*/
/**********************************************************************/
void PrintLongMatrix(
		     long *x,    /* pointer to the matrix of longs */
		     long nrow,  /* number of rows */
		     long ncol   /* number of columns */
		     )
{  
  long i,j;
  for (i=0;i<nrow;i++){
    for(j=0;j<ncol;j++){
      Rprintf("%ld ",x[j*nrow+i]);
    }
    Rprintf("\n");
  }
}

/**********************************************************************/
/*

  PrintIntMatrix

  Print a nrow by ncol matrix of ints to the screen.

*/
/**********************************************************************/
void PrintIntMatrix(
		    int *x,       /* pointer to the matrix of ints */
		    long nrow,    /* number of rows */
		    long ncol     /* number of columns */
		    )
{  
  long i,j;
  for (i=0;i<nrow;i++){
    for(j=0;j<ncol;j++){
      Rprintf("%d ",x[j*nrow+i]);
    }
    Rprintf("\n");
  }
}



/**********************************************************************/
/*

  PrintCharMatrix

  Print a nrow by ncol matrix of chars to the screen.

*/
/**********************************************************************/
void PrintCharMatrix(
		     char *x,     /* pointer to the matrix of chars */
		     long nrow,   /* number of rows */
		     long ncol    /* number of columns */
		     )
{
  long i,j;
  for (i=0;i<nrow;i++){
    for(j=0;j<ncol;j++){
      Rprintf("%c",x[j*nrow+i]); 
    }
  }
}


/**********************************************************************/
/*

  PrintDoubleMatrix2File

  Print a nrow by ncol matrix of doubles to a file.

*/
/**********************************************************************/
void PrintDoubleMatrix2File(
			    FILE *name,    /* file to which to print */
			    double *x,     /* pointer to the matrix of double */
			    long nrow,     /* number of rows */
			    long ncol      /* number of columns */
			    )
{
  long i,j;
  for (i=0;i<nrow;i++){
    for(j=0;j<ncol;j++){
	fprintf(name,"%.3lf ",x[j*nrow+i]);
    }
    fprintf(name,"\n");
  }
}


/**********************************************************************/
/*

  PrintCharMatrix2File

  Print a nrow by ncol matrix of chars to a file.

*/
/**********************************************************************/
void PrintCharMatrix2File(
			  FILE *name,      /* file to which to print */
			  char *x,         /* pointer to the matrix of chars */
			  long nrow,       /* number of rows */
			  long ncol        /* number of columns */
			  )
{
  long i,j;
  for (i=0;i<nrow;i++){
    for(j=0;j<ncol;j++){
      fprintf(name,"%c",x[j*nrow+i]);
    }
    /*fprintf(name,"\n");*/
  }
}

/**********************************************************************/
/*

  PrintIntMatrix2File

  Print a nrow by ncol matrix of doubles to a file.

*/
/**********************************************************************/
void PrintIntMatrix2File(
			 FILE *name,   /* file to which to print */ 
			 int *x,       /* pointer to the matrix of ints */
			 long nrow,    /* number of rows */
			 long ncol     /* number of columns */
			 )
{  
  long i,j;
  for (i=0;i<nrow;i++){
    for(j=0;j<ncol;j++){
      fprintf(name,"%d ",x[j*nrow+i]);
    }
    fprintf(name,"\n");
  }
}

/**********************************************************************/
/*

  PrintLongMatrix2File

  Print a nrow by ncol matrix of longs to a file.

*/
/**********************************************************************/
void PrintLongMatrix2File(
			  FILE *name,    /* file to which to print */
			  long *x,       /* pointer to matrix of longs */
			  long nrow,     /* number of rows */
			  long ncol      /* number of columns */
			  )
{
  long i,j;
  for (i=0;i<nrow;i++){
    for(j=0;j<ncol;j++){
      fprintf(name,"%ld ",x[j*nrow+i]);
    }
    fprintf(name,"\n");
  }
}


/**********************************************************************/
/*
  Convert2Char

  Convert a number 0,1,2,3 to a DNA letter according to the rule

  0 -> A
  1 -> C
  2 -> G
  3 -> T

  Returns the letter.

*/
/**********************************************************************/
char Convert2Char(
		  long num   /* the number to convert */
		  )
{
  char word;
  if (num!=0 && num!=1 && num!=2 && num!=3  && num!=4){
    Rprintf("Convert2Char: Number must be 0,1,2, or 3. Got %ld. Exiting ...\n", num);
    exit(1);
  }

  if (num==0){
    word='A';
  }else if (num==1){
    word='C';
  }else if (num==2){
    word='G';
  }else if (num==3){
    word='T';
  }else{
    word='X';
  }

  return(word);
}

/**********************************************************************/
/*

  Convert2LowChar

  Convert a number 0,1,2,3 to a DNA letter according to the rule

  0 -> a
  1 -> c
  2 -> g
  3 -> t

  Returns the letter.
  

*/
/**********************************************************************/
char Convert2LowChar(
		     long num  /* the number to convert */
		     )
{
  char word;
  if (num!=0 && num!=1 && num!=2 && num!=3 && num!=4){
    Rprintf("Convert2Char: Number must be 0,1,2, or 3. Got %ld. Exiting ...\n", num);
    exit(1);
  }

  if (num==0){
    word='a';
  }else if (num==1){
    word='c';
  }else if (num==2){
    word='g';
  }else if (num==3){
    word='t';
  }else{
    word = 'x';
  }

  return(word);
}


/**********************************************************************/
/*
  Convert2String

  Convert a number 0,1,2,3,4,5 to a DNA letter according to the rule

  0 -> A
  1 -> C
  2 -> G
  3 -> T
  4 -> AT
  5 -> GC

  Returns the letter.

*/
/**********************************************************************/
char *Convert2String(
		    long num   /* the number to convert */
		    ) 
{
  char *word;

  if (num!=0 && num!=1 && num!=2 && num!=3 && num!=4 && num!=5){
    Rprintf("Convert2Char: Number must be 0,1,2,3,4 or 5. Got %ld. Exiting ...\n", num);
    exit(1);
  }
  word = "GC";

  switch(num){
  case 0:
    word = "A ";
    break;
  case 1:
    word = "C ";
    break;
  case 2:
    word = "G ";
    break;
  case 3:
    word = "T ";
    break;
  case 4:
    word = "AT";
    break;
  case 5:
    word = "GC";
  }

  return(word);
}
