/*****************************************************************************
*  			     					             *
*      cosmo - Constrained search for motifs in DNA sequences                *
*      Copyright (C) 2006  Oliver Bembom (bembom@berkeley.edu)               *
*                                                                            *
*      This file is part of cosmo.                                           *
*                                                                            *
*      cosmo is free software; you can redistribute it and/or                *
*      modify it under the terms of the GNU Lesser General Public            *
*      License as published by the Free Software Foundation; either          *
*      version 2.1 of the License, or (at your option) any later version.    *
*                                                                            *
*      cosmo is distributed in the hope that it will be useful,              *
*      but WITHOUT ANY WARRANTY; without even the implied warranty of        *
*      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU     *
*      Lesser General Public License for more details.                       *
*                                                                            *
*      You should have received a copy of the GNU Lesser General Public      *
*      License along with this program; if not, write to the Free Software   *
*      Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,            *
*      MA  02110-1301, USA.                                                  *
*                                                                            *
*                                                                            *
*****************************************************************************/


/*****************************************************************************
*                                                                            *
*      This module contains functions for interacting with donlp2() to       *
*      maximize the likelihood function under constraints.                   *
*      Before calling donlp2() for the first time, a pointer to the dataset  *
*      must be copied to the static variable myDataSet by calling            *
*      init_NLP().                                                           *
*      When optimization routine donlp2() is called, starting value must be  *
*      stored in dataSet->svTheta, validation groups to be excluded must be  *
*      stored in dataSet->excludeCv, and fold for cross-validation must be   *
*      stored in dataSet->fold.                                              *                 
*      To include all sequences, make sure to set fold to 1!                 *
*      On exit, final parameter estimates will be stored in                  * 
*      dataSet->svTheta.                                                     *
*                                                                            *
*****************************************************************************/

#include "o8para.h"
#include <R_ext/PrtUtil.h>
#include <R_ext/Memory.h>
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <float.h>
#include "macros.h"
#include "cosmo.h"
#include "lstarts.h"  
#include "matLib.h"
#include "optimize.h"
#include "theta.h"
  
static DATASET *myDataSet;
static int *liste_loc;

 

/**********************************************************************/
/*
  user_init_size

  Initialize size of the problem by copying information out of
  dataSet->svTheta->conMod.

*/
/**********************************************************************/
void user_init_size(
		    void
		    ) 
{
#define  X extern
#include "o8comm.h"
#include "o8fint.h"
#include "o8cons.h"
#undef   X

  /* problem dimension n = dim(x), nlin=number of linear constraints
     nonlin = number of nonlinear constraints  */
   
  MODEL *mod;

  mod      =  myDataSet->svTheta->conMod; 
  o8n      =  mod->n; 
  o8nlin   =  mod->nlin;
  o8nonlin =  mod->nonlin;
  o8iterma =  30 * o8n; /* changed this one from 30*/
  o8nstep  =  20;

  if (DEBUG>=4) Rprintf("n = %d, nlin = %d, nonlin = %d, mType = %d, intensity = %lf, iterma = %d\n", o8n, o8nlin, o8nonlin, myDataSet->svTheta->mType, myDataSet->svTheta->intensity, o8iterma);
  
}


/**********************************************************************/
/*
  user_init

  Standard setup for donlp2-intv.
  Copy starting values out of dataSet->svTheta.

*/
/**********************************************************************/
void user_init(
	       void
	       ) 
{
#define  X extern
#include "o8comm.h"
#include "o8fint.h"
#include "o8cons.h"
#undef   X
    
  static INTEGER  i,j;   /* indices */
  long pos;              /* position in motif */
  long nuc;              /* nucleotide */
  MODEL *mod;            /* interpreted constraint set */
  THETA *svTheta;        /* starting value */
  int width;             /* width of motif */
  int seqNum;            /* current sequence number */


  svTheta   = myDataSet->svTheta;
  mod       = svTheta->conMod;
  width     = svTheta->width;


  /* name is ident of the example/user and can be set at users will       */
  /* the first static character must be alphabetic. 40 characters maximum */
  strcpy(o8name,"maximOut");
  o8silent = TRUE;

  /* x is initial guess and also holds the current solution */
  /* problem dimension n = dim(x), nlin=number of linear constraints
     nonlin = number of nonlinear constraints  */

  /* analytical gradients for all models except for TCM */
  o8analyt =  (svTheta->mType==TCM) ? FALSE : TRUE;


  o8epsdif = 1.e-16;   /* gradients exact to machine precision */
  /* if you want numerical differentiation being done by donlp2 then:*/
  /* epsfcn   = 1.e-16; *//* function values exact to machine precision */
  o8taubnd   = 1.e-6; 
  /* bounds may be violated at most by taubnd in finite differencing */
  /*  bloc    = TRUE; */
  /* if one wants to evaluate all functions  in an independent process */
  o8difftype = 1; /* the most accurate and most expensive choice */
    
  o8nreset = o8n;
    
  o8del0 = 1e-2;   /* Fiddle with this one! */
  o8tau0 = 1e-2;   /* Fiddle with this one! */
  o8tau  = 0.1e0;

  liste_loc = IntMalloc(mod->n+1);

  /* Fill in starting value */
  i = 1;
  for(pos=0;pos<width;pos++){
    for (nuc=0;nuc<4;nuc++){
      o8x[i] = svTheta->pwm[PWM(pos,nuc)];
      i++;
    }
  }

  /* profile shape parameters */
  for (j=0;j<mod->numProfParms;j++){
    o8x[i] = svTheta->profParms[j];
    if (DEBUG >5) Rprintf("x[%d] is ProfParm[%d] and = %lf\n", i,j,o8x[i]);
    i++;
  }

    
  /*  set lower and upper bounds */
  o8big = 1.e20;
  for (i=0;i<mod->n + mod->nlin + mod->nonlin;i++){
    o8low[i+1] = mod->low[i];
    o8up[i+1]  = mod->up[i];
  }

  if (DEBUG>=4){
    Rprintf("bounds: low, up, start\n");
    PrintDoubleMatrix2File(stderr,o8low,1,mod->n+mod->nlin+mod->nonlin+1);
    Rprintf("\n");
    PrintDoubleMatrix2File(stderr,o8up,1,mod->n+mod->nlin+mod->nonlin+1);
    Rprintf("\n");
    PrintDoubleMatrix2File(stderr,o8x,1,mod->n+mod->nlin+mod->nonlin+1);
    Rprintf("\n");
  }

  /* store coefficients of linear constraints directly in gres */
  for ( i = 1 ; i<= o8nlin ; i++ ){ 
    for  ( j = 1 ; j <= o8n ; j++ ){
      o8gres[j][i] = mod->linConMatrix[LCM(i-1,j-1)];
    }
  }
  
  if (DEBUG>=4){
    Rprintf("user_init: Matrix A:\n");
    for ( i = 1 ; i<= o8nlin ; i++ ){ 
      for  ( j = 1 ; j <= o8n ; j++ ){
	Rprintf("%d ", (int)o8gres[j][i]);
      }
      Rprintf("\n");
    }
  }


  if (DEBUG>=3){
    Rprintf("\nuser_init: excludeCv:\n");
    PrintIntMatrix2File(stderr,myDataSet->excludeCv,1,myDataSet->fold);
    if (DEBUG>=4){
      Rprintf("user_init: skip the following sequences:\n");
      for (seqNum=0; seqNum<myDataSet->numSeqs;seqNum++){
	if (myDataSet->excludeCv[myDataSet->samples[seqNum].cvGroup]) Rprintf("user_init: skip sequence %d\n", seqNum);
      
      }
    }
  }
  
  return;
}

/**********************************************************************/
/*
  setup

  Set special global parameters here.

*/
/**********************************************************************/
void setup(
	   void
	   ) 
{
#define  X extern
#include "o8comm.h"
#undef   X

  /* o8epsphi = 1e-10;*/  /* changed this one */
  o8te0=FALSE;       /* print detailed output? */
  o8te2=TRUE;
  o8te3=TRUE;
  return;
}

/**********************************************************************/
/*
  solchk

  At end of algorithm, copy final parameter estimates to 
  dataSet->svTheta. Check if algorithm converged or not.

*/
/**********************************************************************/
void solchk(
	    void
	    ) 
{
#define  X extern
#include "o8comm.h"
#undef   X
#include "o8cons.h"
   
  int k;          /* an index */
  int ppNum;      /* an index for looping over profile parameters */
  int pos;        /* position in motif */
  int nuc;        /* nucleotide */
  int width;      /* width of motif */
  THETA *svTheta; /* will contain the final parameter estimates on exit */
  MODEL *mod;     /* interpreted constraint set */

  svTheta   = myDataSet->svTheta;  
  width     = svTheta->width;
  mod       = svTheta->conMod;

  k = 1;
  for(pos=0;pos<width;pos++){
    for (nuc=0;nuc<4;nuc++){
      svTheta->pwm[PWM(pos,nuc)] = o8x[k];
      k++;
    }
  }

  /* profile shape parameters */
  for (ppNum=0;ppNum<mod->numProfParms;ppNum++){
    svTheta->profParms[ppNum] = o8x[k];
    k++;
  }

  if (DEBUG>=3){
    Rprintf("\noptimize:\n");
    PrintDoubleMatrix2File(stderr,svTheta->pwm,4,svTheta->width);
  }
 
  /* if no feasible solution is found, algorithm returns o8fx = 0 */
  svTheta->converged = (o8fx!=0) ;
  svTheta->logLR     = -o8fx;

  if (!myDataSet->R) myfree(liste_loc);

  return;
}

/**********************************************************************/
/*
  ef

  Evaluate the objective function and assign value to *o8fx

*/
/**********************************************************************/
void ef(
	DOUBLE o8x[],
	DOUBLE *o8fx
	) 
{
#define  X extern
#include "o8fuco.h"
#undef   X
   

  int k;                /* an index */
  int nuc;              /* nucleotide */
  int pos;              /* position in motif */
  int seqNum;           /* number of current sequence */
  int numSeqs;          /* number of sequences */
  THETA *svTheta;       /* parameter estimates */
  double intensity;     /* intensity parameter */
  int width;            /* motif width */
  MTYPE mType;          /* model type */
  BOOLEAN excludeAny;   /* exclude any validation groups for CV purposes? */
  int *excludeCv;       /* which validation groups to exclude */

  numSeqs     = myDataSet->numSeqs;
  excludeCv   = myDataSet->excludeCv;
  svTheta     = myDataSet->svTheta;
  intensity   = svTheta->intensity;
  width       = svTheta->width;
  mType       = svTheta->mType;

  /* Are we excluding any sequences? */
  excludeAny = FALSE;
  for (k=0;k<myDataSet->fold;k++) excludeAny = excludeAny | excludeCv[k];

  /* copy current parameter estimates into svTheta->pwm */
  k = 1; 
  for(pos=0;pos<width;pos++){
    for (nuc=0;nuc<4;nuc++){
      svTheta->pwm[PWM(pos,nuc)]=o8x[k];
      k++;
    }
  }

  /* Calculate mProbs1 from scratch */
  getNewMProbs(myDataSet,svTheta->pwm,width,1,excludeAny, excludeCv);
  
  /* evaluate objective function */
  *o8fx = 0;    
  for (seqNum=0;seqNum<numSeqs;seqNum++){

    /* skip this sequence if necessary */
    if (myDataSet->fold!=1 && excludeCv[myDataSet->samples[seqNum].cvGroup])  continue;


    *o8fx -= seqLLR(myDataSet, seqNum, mType, intensity, width, 1);
  }
    
}

/**********************************************************************/
/*
  egradf

  Evaluate the gradient of objective function                     
	
  This function assumes that ef() was called earlier so that 
  contributions to likelihood function for each possible motif start
  site are stored in dataSet->seqProbs.

*/
/**********************************************************************/
void egradf(
	    DOUBLE o8x[],
	    DOUBLE o8gradf[]
	    ) 
{
#define  X extern
#include "o8fuco.h"
#undef   X
  
  int k;                /* an index */
  int nuc;              /* nucleotide */
  int pos;              /* position in motif */
  int seqNum;           /* number of current sequence */
  int cutNum;           /* number of current cut */
  int start;            /* motif start position */
  int first;            /* first postion in cut */
  int last;             /* last position in cut */
  int *excludeCv;       /* which validation group is to be excluded to CV purposes? */
  int numSeqs;          /* number of sequences */
  THETA *svTheta;       /* parameter estimates */
  double intensity;     /* intensity parameter */
  int width;            /* motif width */
  MTYPE mType;          /* model type */
  int cutLength;        /* length of cut pieces */
  BOOLEAN revComp;      /* consider motifs on reverse complement? */
  double *posGrad;      /* gradient w.r.t. one entry of the PWM */
  SAMPLE *sample;       /* current sample */
  int *seq;             /* current sequence */
  int *rcSeq;           /* reverse complement of current sequence */
  long seqLength;       /* length of current sequence */
  int numCuts;          /* number of cuts in current sequence */
  double *seqProbs;     /* contribution to likelihood function of each possible motif start site on the forward strand */
  double *rcSeqProbs;   /* contribution to likelihood function of each possible motif start site on the reverse strand */
  double *cutLRs;       /* likelihood ratios for the separate cuts */
  double seqLR;         /* likelihood ratio for the whole sequence */
  double cutFac;        /* factor for integrating out over different possible motif start sites in cut model */
  double seqFac;        /* factor for integrating out over different possible motif start sites */

 
  numSeqs     = myDataSet->numSeqs;
  svTheta     = myDataSet->svTheta;
  intensity   = svTheta->intensity;
  width       = svTheta->width;
  mType       = svTheta->mType;
  revComp     = myDataSet->revComp;
  posGrad     = myDataSet->posGrad;
  excludeCv   = myDataSet->excludeCv;

  /* copy parameter estimates into position weight matrix of svTheta */
  k = 1;
  for(pos=0;pos<width;pos++){
    for (nuc=0;nuc<4;nuc++){
      svTheta->pwm[PWM(pos,nuc)]=o8x[k];
      o8gradf[k] = 0;
      k++;
    }
  }

  if(DEBUG>=3){
    Rprintf("Entering gradf():\n");
    PrintDoubleMatrix2File(stderr,svTheta->pwm,4,svTheta->width);
    for (seqNum=0;seqNum<numSeqs; seqNum++){
      sample = myDataSet->samples + seqNum;
      seqProbs   = sample->seqProbs;
      seqLength  = sample->length;
      seqLR      = sample->seqLR;
      Rprintf("seq %d with seqLR = %lf\n", seqNum, seqLR);
    }
  }

  switch(mType){

  case OOPS:

    /* grad wrt pwm parm for X_i = -(sum_(hasThisNuc) rcFac/numStart *mProb*backLR)/(LR*pwmVal) */
    for (pos=0;pos<width;pos++){
      for (seqNum=0;seqNum<numSeqs;seqNum++){

	sample     = myDataSet->samples + seqNum;
	seq        = sample->seq;
	rcSeq      = sample->rcSeq;
	seqLength  = sample->length;
	seqProbs   = sample->seqProbs;
	rcSeqProbs = sample->rcSeqProbs;
	seqLR      = sample->seqLR;

	/* skip this sequence if necessary */
	if (myDataSet->fold!=1 && excludeCv[myDataSet->samples[seqNum].cvGroup]) continue;
	
	for (nuc=0;nuc<4;nuc++) posGrad[nuc] = 0;
	
	/* factor for integrating over different starting positions and possible reverse complement */
	seqFac  = 1 / ((double) (1 + revComp)*(seqLength - width + 1));
	
	for (start=0;start<seqLength-width+1;start++){
	  nuc             = seq[start + pos];
	  if (nuc<4) posGrad[nuc]   -= seqProbs[start];
	}
	
	if (revComp){
	  for (start=0;start<seqLength-width+1;start++){
	    nuc           = rcSeq[RS(start) + pos];
	    if (nuc<4) posGrad[nuc] -= rcSeqProbs[start];
	  }
	}
	
	for (nuc=0;nuc<4;nuc++) o8gradf[PWM(pos,nuc)+1] += seqFac*posGrad[nuc]/seqLR;
      }
    }
   

    for (pos=0;pos<width;pos++){
      for (nuc=0;nuc<4;nuc++){
	if (svTheta->pwm[PWM(pos,nuc)]!=0) o8gradf[PWM(pos,nuc)+1] /= svTheta->pwm[PWM(pos,nuc)];
      }
    }
    
    break;

  case ZOOPS:

    /* grad wrt pwm parm for X_i = -intensity*(sum_(hasThisNuc) rcFac/numStart *mProb*backLR)/(LR*pwmVal) */
    for (pos=0;pos<width;pos++){
      for (seqNum=0;seqNum<numSeqs;seqNum++){

	sample     = myDataSet->samples + seqNum;
	seq        = sample->seq;
	rcSeq      = sample->rcSeq;
	seqLength  = sample->length;
	seqProbs   = sample->seqProbs;
	rcSeqProbs = sample->rcSeqProbs;
	seqLR      = sample->seqLR;

	/* skip this sequence if necessary */
	if (myDataSet->fold!=1 && excludeCv[myDataSet->samples[seqNum].cvGroup]) continue;
	
	for (nuc=0;nuc<4;nuc++) posGrad[nuc] = 0;
	
	/* factor for integrating over different starting positions and possible reverse complement */
	seqFac  = 1 / ((double) (1 + revComp)*(seqLength - width + 1));
	
	for (start=0;start<seqLength-width+1;start++){
	  nuc             = seq[start + pos];
	  if (nuc<4) posGrad[nuc]   -= seqProbs[start];
	}
	
	if (revComp){
	  for (start=0;start<seqLength-width+1;start++){
	    nuc           = rcSeq[RS(start) + pos];
	    if (nuc<4) posGrad[nuc] -= rcSeqProbs[start];
	  }
	}
	
	for (nuc=0;nuc<4;nuc++) o8gradf[PWM(pos,nuc)+1] += seqFac*posGrad[nuc]/seqLR;
      }
    }
     
    for (pos=0;pos<width;pos++){
      for (nuc=0;nuc<4;nuc++){
	if (svTheta->pwm[PWM(pos,nuc)]!=0) o8gradf[PWM(pos,nuc)+1] *= intensity/svTheta->pwm[PWM(pos,nuc)];
      }
    }
    
    break;


  case OVER:

    cutLength   = myDataSet->cutLengths[width-myDataSet->minWidth];

    /* grad wrt pwm parm for X_i = -intensity*(sum_(hasThisNuc) rcFac/numStart *mProb*backLR)/(LR*pwmVal) */
    for (pos=0;pos<width;pos++){
      for (seqNum=0;seqNum<numSeqs;seqNum++){

	sample     = myDataSet->samples + seqNum;
	seq        = sample->seq;
	rcSeq      = sample->rcSeq;
	seqLength  = sample->length;
	seqProbs   = sample->seqProbs;
	rcSeqProbs = sample->rcSeqProbs;
	numCuts    = sample->numCuts[width-myDataSet->minWidth];
	cutLRs     = sample->cutLRs;

	/* skip this sequence if necessary */
	if (myDataSet->fold!=1 && excludeCv[myDataSet->samples[seqNum].cvGroup]) continue;

	for (cutNum=0;cutNum<numCuts;cutNum++){
		
	  for (nuc=0;nuc<4;nuc++) posGrad[nuc] = 0;

	  /* be careful not to go over the end of the sequence here */		
	  first   = cutNum*cutLength;
	  last    = MIN((cutNum+1)*cutLength-1, seqLength - width);

	  if (last < first) continue;

	  /* factor for integrating over different starting positions and possible reverse complement */
	  cutFac  = 1 / ((double) (1 + revComp)*(last - first + 1));

	  for (start=first;start<=last;start++){
	    nuc             = seq[start + pos];
	    if (nuc<4) posGrad[nuc]   -= seqProbs[start];
	  }

	  if (revComp){
	    for (start=first;start<=last;start++){
	      nuc           = rcSeq[RS(start) + pos];
	      if (nuc<4) posGrad[nuc] -= rcSeqProbs[start];
	    }
	  }
		
	  for (nuc=0;nuc<4;nuc++) o8gradf[PWM(pos,nuc)+1] += cutFac*posGrad[nuc]/cutLRs[cutNum];
				
	}
      }
    }

    for (pos=0;pos<width;pos++){
      for (nuc=0;nuc<4;nuc++){
	if (svTheta->pwm[PWM(pos,nuc)]!=0) o8gradf[PWM(pos,nuc)+1] *= intensity/svTheta->pwm[PWM(pos,nuc)];
      }
    }

    break;

  case CUT:

    cutLength   = myDataSet->cutLengths[width-myDataSet->minWidth];

    /* grad wrt pwm parm for X_i = -intensity*(sum_(hasThisNuc) rcFac/numStart *mProb*backLR)/(LR*pwmVal) */
    for (pos=0;pos<width;pos++){
      for (seqNum=0;seqNum<numSeqs;seqNum++){

	sample     = myDataSet->samples + seqNum;
	seq        = sample->seq;
	rcSeq      = sample->rcSeq;
	seqLength  = sample->length;
	seqProbs   = sample->seqProbs;
	rcSeqProbs = sample->rcSeqProbs;
	numCuts    = sample->numCuts[width-myDataSet->minWidth];
	cutLRs     = sample->cutLRs;

	/* skip this sequence if necessary */
	if (myDataSet->fold!=1 && excludeCv[myDataSet->samples[seqNum].cvGroup]) continue;

	for (cutNum=0;cutNum<numCuts;cutNum++){
		
	  for (nuc=0;nuc<4;nuc++) posGrad[nuc] = 0;

	  /* be careful not to go over the end of the sequence here */		
	  first   = cutNum*cutLength;
	  last    = MIN((cutNum+1)*cutLength - width + 1, seqLength - width);

	  if (last < first) continue;

	  /* factor for integrating over different starting positions and possible reverse complement */
	  cutFac  = 1 / ((double) (1 + revComp)*(last - first + 1));

	  for (start=first;start<=last;start++){
	    nuc             = seq[start + pos];
	    if (nuc<4) posGrad[nuc]   -= seqProbs[start];
	  }

	  if (revComp){
	    for (start=first;start<=last;start++){
	      nuc           = rcSeq[RS(start) + pos];
	      if (nuc<4) posGrad[nuc] -= rcSeqProbs[start];
	    }
	  }
		
	  for (nuc=0;nuc<4;nuc++) o8gradf[PWM(pos,nuc)+1] += cutFac*posGrad[nuc]/cutLRs[cutNum];
				
	}
      }
    }

    for (pos=0;pos<width;pos++){
      for (nuc=0;nuc<4;nuc++){
	if (svTheta->pwm[PWM(pos,nuc)]!=0) o8gradf[PWM(pos,nuc)+1] *= intensity/svTheta->pwm[PWM(pos,nuc)];
      }
    }

    break;

  default:
    Rprintf("gradf: mType %d not recognized. Exiting ...\n", mType);
  }


}

/**********************************************************************/
/*
  econ

  Compute the i-th equality constaint, value is hxi                 
	
*/
/**********************************************************************/
void econ(
	  INTEGER type,
	  INTEGER liste[], 
	  DOUBLE o8x[],
	  DOUBLE con[], 
	  LOGICAL err[]
	  ) 
{
#define  X extern
#include "o8fuco.h"
#undef   X

  int i,j,k;            /* various indices */
  int pos;              /* position in motif */
  int nuc;              /* nucleotide */
  int start;            /* motif start site */
  double prob;          /* probablity */
  int width;            /* width of motif */
  MODEL *mod;           /* interpreted constraints */
  double *pwm;          /* position weight matrix */
  double logTwo;        /* constant: log(2) */
  double eps;           /* small error tolerance */
  double subMin;        /* minimum submotif penalty for any window so far */
  double subScore;      /* submotif penalty for this window so far */
  double penalty;       /* submotif penalty for one position */
   


  width       = myDataSet->svTheta->width;
  mod         = myDataSet->svTheta->conMod;
  pwm         = myDataSet->svTheta->pwm;
  logTwo      = log(2);
  eps         = 1e-8;

  k = 1;
  for(pos=0;pos<width;pos++){
    for (nuc=0;nuc<4;nuc++){
      pwm[PWM(pos,nuc)]=o8x[k];
      k++;
    }
  }

  if (type==1){
    liste_loc[0] = mod->nonlin;
    if (DEBUG>4) Rprintf("liste_loc[0] = %d\n", liste_loc[0]);
    for (i=1;i<=mod->nonlin;i++){
      liste_loc[i] = i;
    }
  }else{
    liste_loc[0] = liste[0];
    for (i=1;i<=liste[0];i++){
      liste_loc[i] = liste[i];
    }
  }

  for (j=1;j<=liste_loc[0];j++){
    i = liste_loc[j] - 1; /*back to zero-based arrays! */
    if (DEBUG>4) Rprintf("econ: index i = %d\n",i);

    con[i+1] = 0;
    switch(mod->nlCons[i].pCon.conType){
    case bdd:

      /* 2 + sum(f_i*log(f_i)/log(2)) */
      for (nuc=0;nuc<4;nuc++){
	prob = pwm[PWM(mod->nlCons[i].pCon.pos,nuc)];
	if (prob<eps) continue;
	con[i+1] += log(prob)*prob;
      }
      con[i+1] /= logTwo;
      con[i+1] += 2;
      break;

    case lin:
      /* 2 + sum(f_i*log(f_i)/log(2)) - (parm1 + (parm2-parm1)*xPos) */
      for (nuc=0;nuc<4;nuc++){
	prob = pwm[PWM(mod->nlCons[i].pCon.pos,nuc)];
	if (prob<eps) continue;
	con[i+1] += log(prob)*prob;
      }
      con[i+1] /= logTwo;
      con[i+1] += 2;
      con[i+1] -= o8x[mod->nlCons[i].pCon.parm1 + 1] + (o8x[mod->nlCons[i].pCon.parm2 + 1] - o8x[mod->nlCons[i].pCon.parm1 + 1])*mod->nlCons[i].pCon.xPos;

      break;

    case sub:
      subMin = mod->nlCons[i].subCon.length;
      for (start=0;start< mod->width - mod->nlCons[i].subCon.length + 1; start++){
	subScore = 0;
	for (pos = 0; pos< mod->nlCons[i].subCon.length;pos++){
	  /* penalty = exp(-5*prob) */
	  penalty = exp(-5*pwm[PWM(start+pos,mod->nlCons[i].subCon.motif[pos])]);
	  if (DEBUG>=3) Rprintf("Adding penalty %lf for freq %lf\n",
			       penalty, pwm[PWM(start+pos,mod->nlCons[i].subCon.motif[pos])]);
	  subScore += penalty;
	}
	if (DEBUG>=3) Rprintf("subScore for position %d: %lf\n", start, subScore);
	subMin = MIN(subMin,subScore);
      }
      if (DEBUG>=3) Rprintf("subMin = %lf\n", subMin);
      con[i+1] = subMin;		
      break;

    default:
      for (nuc=0;nuc<4;nuc++){
	prob = pwm[PWM(mod->nlCons[i].pCon.pos,nuc)];
	if (prob<eps) continue;
	con[i+1] += log(prob)*prob;
      }	

      switch(mod->nlCons[i].pCon.ref){
      case prevPos:
	/* sum_i(f_ij*log(f_ij)/log(2)) - sum_i(f_i(j-1)*log(f_i(j-1))/log(2)) */
	for (nuc=0;nuc<4;nuc++){
	  prob = pwm[PWM(mod->nlCons[i].pCon.pos-1,nuc)];
	  if (prob<eps) continue;
	  con[i+1] -= log(prob)*prob;
	}
	con[i+1] /= logTwo;
	break;

      case prevBrk:
	/* 2 + sum_i(f_ij*log(f_ij)/log(2)) - parm1 */
	con[i+1] /= logTwo;
	con[i+1] += 2 - o8x[mod->nlCons[i].pCon.parm1 + 1];
	break;

      default:
	/* parm2 - (2 + sum_i(f_ij*log(f_ij)/log(2))) */
	con[i+1] /= logTwo;
	con[i+1] = o8x[mod->nlCons[i].pCon.parm2 + 1] -(con[i+1] + 2);
      }
    }
    
    if (DEBUG>4) Rprintf("con[%d] = %lf\n",i+1,con[i+1]);
  }

  return;
}

/**********************************************************************/
/*
  econgrad

  Compute the gradient of the i-th equality constraint           
	
*/
/**********************************************************************/
void econgrad(
	      INTEGER liste[] ,
	      INTEGER shift ,  
	      DOUBLE o8x[],
	      DOUBLE **grad
	      ) 
{
#define  X extern
#include "o8fuco.h"
#undef   X
   
  int i,j,k;            /* various indices */
  int pos;              /* position in motif */
  int nuc;              /* nucleotide */
  int start;            /* motif start site */
  double prob;          /* probablity */
  int width;            /* width of motif */
  MODEL *mod;           /* interpreted constraints */
  double *pwm;          /* position weight matrix */
  double logTwo;        /* constant: log(2) */
  double eps;           /* small error tolerance */
  double small;         /* a small number */

  int minStart;         /* start site for submotif that incurs minimum penalty */
  double subMin;        /* minimum submotif penalty for any window so far */
  double subScore;      /* submotif penalty for this window so far */
  double penalty;       /* submotif penalty for one position */


  width       = myDataSet->svTheta->width;
  mod         = myDataSet->svTheta->conMod;
  pwm         = myDataSet->svTheta->pwm;
  logTwo      = log(2);
  eps         = 1e-8;
  small       = 1e-20;

  k = 1;
  for(pos=0;pos<width;pos++){
    for (nuc=0;nuc<4;nuc++){
      pwm[4*pos + nuc]=o8x[k];
      k++;
    }
  }

  liste_loc[0] = liste[0];
  for (i=1;i<=liste[0];i++){
    liste_loc[i] = liste[i];
  }
    
  for (j=1;j<=liste_loc[0];j++){
    i = liste_loc[j] - 1; /*back to zero-based arrays!*/
    if (DEBUG>4) Rprintf("econgrad: index i = %d\n",i);

    /* Initialize gradient to zero */
    for (k=1;k<=mod->n;k++){
      grad[k][i+1+shift] = 0.0;
    }

	
    switch(mod->nlCons[i].pCon.conType){
    case bdd:
      /* grad wrt pwm parm at X_i = (log(pwmParm) + 1)/log(2)*/
      for (nuc=0;nuc<4;nuc++){
	prob = pwm[4*mod->nlCons[i].pCon.pos + nuc];
	grad[4*mod->nlCons[i].pCon.pos + nuc + 1][i+1+shift] = prob<eps ? small : (log(prob) + 1)/logTwo;
      }
      break;

    case lin:
      for (nuc=0;nuc<4;nuc++){
	prob = pwm[4*mod->nlCons[i].pCon.pos + nuc];
	grad[4*mod->nlCons[i].pCon.pos + nuc + 1][i+1+shift] = prob<eps ? small : (log(prob) + 1)/logTwo;
      }
      grad[mod->nlCons[i].pCon.parm1 + 1][i+1+shift] = -1 + mod->nlCons[i].pCon.xPos; 
      grad[mod->nlCons[i].pCon.parm2 + 1][i+1+shift] = -mod->nlCons[i].pCon.xPos; 
      break;

    case sub:

      /* First find in which window the minimum occurs */
      subMin = mod->nlCons[i].subCon.length;
      minStart = 0;
      for (start=0;start<mod->width - mod->nlCons[i].subCon.length + 1; start++){
	subScore = 0;
	if (DEBUG>=3) Rprintf("egrad: start = %d\n", start);
	for (pos = 0; pos<mod->nlCons[i].subCon.length;pos++){
	  penalty = exp(-5*pwm[PWM(start+pos,mod->nlCons[i].subCon.motif[pos])]);
	  if (DEBUG>=3) Rprintf("egrad: penalty %lf for freq %lf\n",
			       penalty, pwm[PWM(start+pos,mod->nlCons[i].subCon.motif[pos])]);
	  subScore += penalty;
	}
	if (subScore<subMin){
	  subMin   = subScore;
	  minStart = start;
	}
      }

      /* Now get gradient as grad = -5*exp(-5*prob) */
      for (pos = 0; pos<mod->nlCons[i].subCon.length;pos++){
	grad[4*(minStart+pos) + mod->nlCons[i].subCon.motif[pos] + 1][i+1+shift] = 
	  -5*exp(-5*pwm[PWM(minStart+pos,mod->nlCons[i].subCon.motif[pos])]);
      }
      break;
		
    default:
      for (nuc=0;nuc<4;nuc++){
	prob = pwm[4*mod->nlCons[i].pCon.pos + nuc];
	grad[4*mod->nlCons[i].pCon.pos + nuc + 1][i+1+shift] = prob<eps ? small : (log(prob) + 1)/logTwo;
      }
		 
      switch(mod->nlCons[i].pCon.ref){
      case prevPos:
	for (nuc=0;nuc<4;nuc++){
	  prob = pwm[4*(mod->nlCons[i].pCon.pos-1) + nuc];
	  grad[4*(mod->nlCons[i].pCon.pos-1) + nuc + 1][i+1+shift] = prob<eps ? small : - (log(prob) + 1 )/logTwo;
	}
	break;

      case prevBrk:
	grad[mod->nlCons[i].pCon.parm1 + 1][i+1+shift] = -1; 
	break;

      default:
	for (nuc=0;nuc<4;nuc++){
	  grad[4*mod->nlCons[i].pCon.pos + nuc + 1][i+1+shift] = -grad[4*mod->nlCons[i].pCon.pos + nuc + 1][i+1+shift]; 		
	}
	grad[mod->nlCons[i].pCon.parm2 + 1][i+1+shift] = 1; 
      }
    }
	
    if (DEBUG>4){
      for (k=0;k<mod->n;k++){
	Rprintf("%5.3lf ",grad[k+1][i+1+shift]);
      }
      Rprintf("\n");
    }

  }
	    
  return;
}

/**********************************************************************/
/*
  eval_extern

  User functions that is called if bloc==TRUE                     
	
*/
/**********************************************************************/
void eval_extern(
		 INTEGER mode
		 ) 
{
  return;
}



/**********************************************************************/
/*
  initNLP

  Get a pointer to the dataset and store in the static variable 
  myDataSet.
	
*/
/**********************************************************************/
void initNLP(
	     DATASET *dataSet  /* the dataset */
	     )
{
  myDataSet = dataSet;
}
