/*****************************************************************************
*    			     					             *
*      cosmo - Constrained search for motifs in DNA sequences                *
*      Copyright (C) 2006  Oliver Bembom (bembom@berkeley.edu)               *
*                                                                            *
*      This file is part of cosmo.                                           *
*                                                                            *
*      cosmo is free software; you can redistribute it and/or                *
*      modify it under the terms of the GNU Lesser General Public            *
*      License as published by the Free Software Foundation; either          *
*      version 2.1 of the License, or (at your option) any later version.    *
*                                                                            *
*      cosmo is distributed in the hope that it will be useful,              *
*      but WITHOUT ANY WARRANTY; without even the implied warranty of        *
*      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU     *
*      Lesser General Public License for more details.                       *
*                                                                            *
*      You should have received a copy of the GNU Lesser General Public      *
*      License along with this program; if not, write to the Free Software   *
*      Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,            *
*      MA  02110-1301, USA.                                                  *
*                                                                            *
*                                                                            *
*****************************************************************************/


/*****************************************************************************
*                                                                            *
*      This module contains various functions for computing posterior        * 
*      probabilities                                                         *
*                                                                            *
*****************************************************************************/

#include <R_ext/PrtUtil.h>
#include <R_ext/Memory.h>
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <float.h>
#include "macros.h"
#include "cosmo.h"
#include "lstarts.h"
#include "matLib.h"
#include "posterior.h"


/**********************************************************************/
/*
  getPostProbs

  Get posterior probabilities for a motif starting in each of the 
  eligible positions given the parameter estimates.

*/
/**********************************************************************/
int getPostProbs(
		  THETA *theta,     /* parameter estimates */
		  DATASET *dataSet  /* the dataset */
		  )
{

  int seqNum;            /* number of current sequence */
  int numSeqs;           /* number of sequences */
  long maxSeqLength;     /* maximum sequence length */
  int minWidth;          /* minimum motif width */
  long seqLength;        /* length of current sequence */
  int revComp;           /* consider reverse complement? */
  int cutNum;            /* index of current cut */
  int numCuts;           /* number of cuts in this sequence */
  int cutLength;         /* length of cut pieces */
  int first;             /* first position in this cut */ 
  int last;              /* last position in this cut */
  SAMPLE *sample;        /* the current sample */
  int *seq;              /* the current sequence */
  double pi;             /* intensity parameter in the ZOOPS model */
  double lambda;         /* intensity parameter in the TCM model */
  double *mProbs;        /* motif probabilities on forward strand */
  double *rcmProbs;      /* motif probabilities on reverse strand */
  double *seqProbs;      /* sequence probabilities */
  double *cumProb;       /* cumulative probabilities up to current position */
  double *cumProbBack;   /* cumulative probabilities under background model */
  double *backLR;        /* contributions of background to likelihood ratio */
  POSTPROB *postProbs;   /* posterior probabilities */
  int start;             /* position in sequence */
  double back;           /* likelihood under background model */
  double seqLR;          /* likelihood ratio for this sequence */
  double cutLR;          /* likelihood ratio for this cut */
  double *pwm;           /* position weight matrix */
  int width;             /* width of motif */
  MTYPE mType;           /* model type */

  pwm           = theta->pwm;
  width         = theta->width;  
  mType         = theta->mType;
  maxSeqLength  = dataSet->maxSeqLength;
  minWidth      = dataSet->minWidth;
  numSeqs       = dataSet->numSeqs;
  revComp       = dataSet->revComp;
  cumProb       = dataSet->cumProb;
  postProbs     = dataSet->postProbs;
  cutLength     = dataSet->cutLengths[width-minWidth];


  /* calculate motif probabilities from scratch */
  getNewMProbs(dataSet, pwm, width, 1,FALSE, NULL);


  switch (mType){
    
  case OOPS:

    for (seqNum=0; seqNum<numSeqs; seqNum++){
      sample        = dataSet->samples + seqNum;
      seq           = sample->seq;
      seqLength     = sample->length;
      backLR        = sample->backLR;
      mProbs        = sample->mProbs1;
      rcmProbs      = sample->rcmProbs1;
      seqProbs      = sample->seqProbs;

      seqLR = 0;
      
      if (revComp){
	for (start=0;start<seqLength-width+1;start++){
	  seqProbs[start] = 0.5*(mProbs[start] + rcmProbs[RS(start)])* backLR[(width-minWidth)*seqLength + start]; 	  
	  seqLR          += seqProbs[start];
	}
      }else{
	for (start=0;start<seqLength-width+1;start++){
	  seqProbs[start]  = mProbs[start] * backLR[(width-minWidth)*seqLength + start];
	  seqLR           += seqProbs[start];
	}
      }
    
      seqLR /= seqLength - width + 1;

      for (start=0;start<seqLength-width+1;start++){
	postProbs[PP(seqNum,start)].seqNum  = seqNum;
	postProbs[PP(seqNum,start)].start   = start;
	postProbs[PP(seqNum,start)].prob    = (seqProbs[start]/(seqLength - width + 1))/seqLR;
	postProbs[PP(seqNum,start)].revComp = (revComp && (rcmProbs[RS(start)] > mProbs[start])) ? 1 : 0;
      }

      for (start=seqLength-width+1;start<seqLength;start++){
	postProbs[PP(seqNum,start)].seqNum  = seqNum;
	postProbs[PP(seqNum,start)].start   = start;
	postProbs[PP(seqNum,start)].prob    = 0;
	postProbs[PP(seqNum,start)].revComp = 0;
      }

      for (start=seqLength;start<maxSeqLength;start++){
	postProbs[PP(seqNum,start)].seqNum  = seqNum;
	postProbs[PP(seqNum,start)].start   = -1;
	postProbs[PP(seqNum,start)].prob    = 0;
	postProbs[PP(seqNum,start)].revComp = 0;
      }

    }

    return(1);

  case ZOOPS:

    pi = theta->intensity;

    for (seqNum=0; seqNum<numSeqs; seqNum++){
      sample        = dataSet->samples + seqNum;
      seq           = sample->seq;
      seqLength     = sample->length;
      backLR        = sample->backLR;
      mProbs        = sample->mProbs1;
      rcmProbs      = sample->rcmProbs1; 
      seqProbs      = sample->seqProbs;

      seqLR = 0;
      
      if (revComp){
	for (start=0;start<seqLength-width+1;start++){
	  seqProbs[start]  = 0.5*(mProbs[start] + rcmProbs[RS(start)])* backLR[(width-minWidth)*seqLength + start]; 
	  seqLR           += seqProbs[start];
	}
      }else{
	for (start=0;start<seqLength-width+1;start++){
	  seqProbs[start]  = mProbs[start] * backLR[(width-minWidth)*seqLength + start];
	  seqLR           += seqProbs[start];
	}
      }

      seqLR = pi/(seqLength - width + 1) * seqLR + (1 - pi);
 
      for (start=0;start<seqLength-width+1;start++){
	postProbs[PP(seqNum,start)].seqNum  = seqNum;
	postProbs[PP(seqNum,start)].start   = start;
	postProbs[PP(seqNum,start)].prob    = pi/(seqLength - width + 1) * seqProbs[start]/seqLR;
	postProbs[PP(seqNum,start)].revComp = (revComp && (rcmProbs[RS(start)] > mProbs[start])) ? 1 : 0;
      }

      for (start=seqLength-width+1;start<seqLength;start++){
	postProbs[PP(seqNum,start)].seqNum  = seqNum;
	postProbs[PP(seqNum,start)].start   = start;
	postProbs[PP(seqNum,start)].prob    = 0;
	postProbs[PP(seqNum,start)].revComp = 0;
      }

      for (start=seqLength;start<maxSeqLength;start++){
	postProbs[PP(seqNum,start)].seqNum  = seqNum;
	postProbs[PP(seqNum,start)].start   = -1;
	postProbs[PP(seqNum,start)].prob    = 0;
	postProbs[PP(seqNum,start)].revComp = 0;
      }

    }

    return(1);

  case TCM:

    lambda       = theta->intensity;
    cumProb      = dataSet->cumProb;

    for (seqNum=0; seqNum<numSeqs; seqNum++){
      sample        = dataSet->samples + seqNum;
      seq           = sample->seq;
      seqLength     = sample->length;
      backLR        = sample->backLR;
      mProbs        = sample->mProbs1;
      rcmProbs      = sample->rcmProbs1; 
      cumProbBack   = sample->cumProbBack;


      for (start=0;start<seqLength+1;start++){
	cumProb[start] = 0;
      }
      
      cumProb[0] = 1;
      
      if (revComp){
	for (start=0;start<seqLength - width + 1;start++){
	  cumProb[start+1] += cumProb[start] * (1 - lambda);
	  
	  back = cumProbBack[CPB(width,start)];
	  cumProb[start+width] += 0.5*cumProb[start]*lambda*(mProbs[start]+rcmProbs[RS(start)])/back;		
	}
      }else{
	for (start=0;start<seqLength - width + 1;start++){
	  cumProb[start+1] += cumProb[start] * (1 - lambda);
	  
	  back = cumProbBack[CPB(width,start)];
	  cumProb[start+width] += cumProb[start]*lambda*mProbs[start]/back;		
	}
      }
      
      for (start=seqLength - width + 1;start<seqLength;start++){
	cumProb[start+1] += cumProb[start];
      }
      

      seqLR = cumProb[seqLength];

      for (start=0;start<seqLength-width+1;start++){
	postProbs[PP(seqNum,start)].seqNum  = seqNum;
	postProbs[PP(seqNum,start)].start   = start;
	postProbs[PP(seqNum,start)].prob    = getCondTcmLik(start,seqNum,lambda,width,dataSet)/seqLR;
	postProbs[PP(seqNum,start)].revComp = (revComp && (rcmProbs[RS(start)] > mProbs[start])) ? 1 : 0;
      }

      for (start=seqLength-width+1;start<seqLength;start++){
	postProbs[PP(seqNum,start)].seqNum  = seqNum;
	postProbs[PP(seqNum,start)].start   = start;
	postProbs[PP(seqNum,start)].prob    = 0;
	postProbs[PP(seqNum,start)].revComp = 0;
      }

      for (start=seqLength;start<maxSeqLength;start++){
	postProbs[PP(seqNum,start)].seqNum  = seqNum;
	postProbs[PP(seqNum,start)].start   = -1;
	postProbs[PP(seqNum,start)].prob    = 0;
	postProbs[PP(seqNum,start)].revComp = 0;
      }


    }

    return(1);


  case OVER:

    pi       = theta->intensity;

    for (seqNum=0; seqNum<numSeqs; seqNum++){
      sample        = dataSet->samples + seqNum;
      seq           = sample->seq;
      seqLength     = sample->length;
      backLR        = sample->backLR;
      mProbs        = sample->mProbs1;
      rcmProbs      = sample->rcmProbs1; 
      seqProbs      = sample->seqProbs;
      numCuts       = sample->numCuts[width-minWidth];

      for (cutNum=0;cutNum <numCuts;cutNum++){
	
	/* cutLR here is the probability over one cut sequence */
	cutLR = 0;
	
	first = cutNum*cutLength;
	
	/* be careful not to go over the end of the sequence here */		    
	last = MIN((cutNum+1)*cutLength-1, seqLength - width + 1);
 
	/* changed this on 1/13/06 */
	/* if cutLength < width, we can't consider the last few cuts */	    
	if (last<first) continue;
	
	if (revComp){
	  for (start=first; start <= last; start++){
	    seqProbs[start]   = 0.5*(mProbs[start] + rcmProbs[RS(start)])* backLR[(width-minWidth)*seqLength + start]; 
	    cutLR            += seqProbs[start]; 
	  }
	}else{
	  for (start=first; start <= last; start++){
	    seqProbs[start]   = mProbs[start]* backLR[(width-minWidth)*seqLength + start];
	    cutLR            += seqProbs[start];        
	  }
	}
	
	cutLR  = pi/(last - first + 1) * cutLR + (1 - pi);
	
	for (start=first;start<=last;start++){
	  postProbs[PP(seqNum,start)].seqNum  = seqNum;
	  postProbs[PP(seqNum,start)].start   = start;
	  postProbs[PP(seqNum,start)].prob    = pi/(last - first + 1) * seqProbs[start]/cutLR;
	  postProbs[PP(seqNum,start)].revComp = (revComp && (rcmProbs[RS(start)] > mProbs[start])) ? 1 : 0;	  
	}

	if (last == seqLength - width + 1){
	  for (start=seqLength-width+1;start<seqLength;start++){
	    postProbs[PP(seqNum,start)].seqNum  = seqNum;
	    postProbs[PP(seqNum,start)].start   = start;
	    postProbs[PP(seqNum,start)].prob    = 0;
	    postProbs[PP(seqNum,start)].revComp = 0;
	  }

	  for (start=seqLength;start<maxSeqLength;start++){
	    postProbs[PP(seqNum,start)].seqNum  = seqNum;
	    postProbs[PP(seqNum,start)].start   = -1;
	    postProbs[PP(seqNum,start)].prob    = 0;
	    postProbs[PP(seqNum,start)].revComp = 0;
	  }

	}
      }
    }

    return(1);

  case CUT:

    pi       = theta->intensity;

    for (seqNum=0; seqNum<numSeqs; seqNum++){
      sample        = dataSet->samples + seqNum;
      seq           = sample->seq;
      seqLength     = sample->length;
      backLR        = sample->backLR;
      mProbs        = sample->mProbs1;
      rcmProbs      = sample->rcmProbs1; 
      seqProbs      = sample->seqProbs;
      numCuts       = sample->numCuts[width-minWidth];

      for (cutNum=0;cutNum <numCuts;cutNum++){
	
	/* cutLR here is the probability over one cut sequence */
	cutLR = 0;
	
	first = cutNum*cutLength;
	
	/* be careful not to go over the end of the sequence here */		    
	last = MIN((cutNum+1)*cutLength - width + 1, seqLength - width + 1);	
 
	/* changed this on 1/13/06 */
	/* if cutLength < width, we can't consider the last few cuts */	    
	if (last<first) continue;
	
	if (revComp){
	  for (start=first; start <= last; start++){
	    seqProbs[start]   = 0.5*(mProbs[start] + rcmProbs[RS(start)])* backLR[(width-minWidth)*seqLength + start]; 
	    cutLR            += seqProbs[start]; 
	  }
	}else{
	  for (start=first; start <= last; start++){
	    seqProbs[start]   = mProbs[start]* backLR[(width-minWidth)*seqLength + start];
	    cutLR            += seqProbs[start];        
	  }
	}
	
	cutLR  = pi/(last - first + 1) * cutLR + (1 - pi);
	
	for (start=first;start<=last;start++){
	  postProbs[PP(seqNum,start)].seqNum  = seqNum;
	  postProbs[PP(seqNum,start)].start   = start;
	  postProbs[PP(seqNum,start)].prob    = pi/(last - first + 1) * seqProbs[start]/cutLR;
	  postProbs[PP(seqNum,start)].revComp = (revComp && (rcmProbs[RS(start)] > mProbs[start])) ? 1 : 0;	  
	}

	if (last == seqLength - width + 1){
	  for (start=seqLength-width+1;start<seqLength;start++){
	    postProbs[PP(seqNum,start)].seqNum  = seqNum;
	    postProbs[PP(seqNum,start)].start   = start;
	    postProbs[PP(seqNum,start)].prob    = 0;
	    postProbs[PP(seqNum,start)].revComp = 0;
	  }

	  for (start=seqLength;start<maxSeqLength;start++){
	    postProbs[PP(seqNum,start)].seqNum  = seqNum;
	    postProbs[PP(seqNum,start)].start   = -1;
	    postProbs[PP(seqNum,start)].prob    = 0;
	    postProbs[PP(seqNum,start)].revComp = 0;
	  }

	}
      }
    }

    return(1);
    
  default:
    Rprintf("getPostProbs: mType %d not recognized. Exiting ...\n", mType);
    return(0);
  }

}
 
/**********************************************************************/
/*
	getCondTcmLik

	Calculate likelihood ratio of one sequence under TCM model, 
	assuming that a motif starts at position motifStart

*/
/**********************************************************************/
double getCondTcmLik(
		     int motifStart,    /* motif start position */
		     int seqNum,        /* sequence for which to calculate conditional likelihood */
		     double lambda,     /* intensity parameter */
		     int width,         /* width of motif */
		     DATASET *dataSet   /* the dataset */
 		     )
{
  
  long maxSeqLength;     /* maximum sequence length */
  int minWidth;          /* minimum motif width */
  BOOLEAN revComp;       /* consider reverse complement? */
  SAMPLE *sample;        /* the current sample */
  int *seq;              /* the current sequence */
  long seqLength;        /* length of current sequence */
  double *mProbs;        /* motif probabilities on forward strand */
  double *rcmProbs;      /* motif probabilities on reverse strand */
  double *cumProb;       /* cumulative probabilities up to current position */
  double *cumProbBack;   /* cumulative probabilities under background model */
  int start;             /* position in sequence */
  double back;           /* likelihood under background model */


  maxSeqLength  = dataSet->maxSeqLength;
  minWidth      = dataSet->minWidth;
  revComp       = dataSet->revComp;
  cumProb       = dataSet->cumProb;

  sample        = dataSet->samples + seqNum;
  seq           = sample->seq;
  seqLength     = sample->length;
  mProbs        = sample->mProbs1;
  rcmProbs      = sample->rcmProbs1; 
  cumProbBack   = sample->cumProbBack;

  for (start=0;start<seqLength+1;start++){
    cumProb[start] = 0;
  }

  cumProb[0] = 1;

  if (revComp){

    for (start=0;start<motifStart - width + 1;start++){
      
      cumProb[start+1] += cumProb[start]*(1-lambda);
      
      back = cumProbBack[CPB(width,start)];
      cumProb[start+width] += 0.5*cumProb[start]*lambda*(mProbs[start]+rcmProbs[RS(start)])/back;				
    }

    for (start=MAX(0, motifStart-width+1);start<motifStart;start++){
      cumProb[start+1] += cumProb[start]*(1-lambda);
    }

    back = cumProbBack[CPB(width,motifStart)];
    cumProb[motifStart+width] += 0.5*cumProb[motifStart]*lambda*(mProbs[motifStart]+rcmProbs[RS(motifStart)])/back;	 

    for (start=motifStart+width;start<seqLength - width + 1;start++){
      
      cumProb[start+1] += cumProb[start]*(1-lambda);
      
      back = cumProbBack[CPB(width,start)];
      cumProb[start+width] += 0.5*cumProb[start]*lambda*(mProbs[start]+rcmProbs[RS(start)])/back;				
    }
    
    for (start=MAX(seqLength - width + 1,motifStart+width);start<seqLength;start++){
      cumProb[start+1] += cumProb[start];
    }

  }else{

    for (start=0;start<motifStart - width + 1;start++){
      
      cumProb[start+1] += cumProb[start]*(1-lambda);
      
      back = cumProbBack[CPB(width,start)];
      cumProb[start+width] += cumProb[start]*lambda*mProbs[start]/back;		
    }

    for (start=MAX(0, motifStart-width+1);start<motifStart;start++){
      cumProb[start+1] += cumProb[start]*(1-lambda);
    }
    
    back = cumProbBack[CPB(width,motifStart)];
    cumProb[motifStart+width] += cumProb[motifStart]*lambda*mProbs[motifStart]/back; 

    for (start=motifStart+width;start<seqLength - width + 1;start++){
      
      cumProb[start+1] += cumProb[start]*(1-lambda);
      
      back = cumProbBack[CPB(width,start)];
      cumProb[start+width] += cumProb[start]*lambda*mProbs[start]/back;				
    }
    
    for (start=MAX(seqLength - width + 1,motifStart+width);start<seqLength;start++){
      cumProb[start+1] += cumProb[start];
    }
  }

  return cumProb[seqLength];
    
}



/**********************************************************************/
/*
   compPostProbs

   Compare to postProbs structs on the basis of their posterior
   probabilities

*/
/**********************************************************************/
int compPostProbs(
		  const void *x,   /* first posterior probability struct */
		  const void *y    /* second posterior probability struct */
		 )
{

  POSTPROB *p1, *p2;

  p1 = (POSTPROB *) x;
  p2 = (POSTPROB *) y;

  if (p1->prob > p2->prob) return -1;
  if (p2->prob > p1->prob) return 1;
  return 0;
}



/**********************************************************************/
/*
   sortPostProbs

   Sort posterior probabilities, either over all sequences or separately
   for each sequence.

*/
/**********************************************************************/
int sortPostProbs(
		   DATASET *dataSet,   /* the dataset */
		   BOOLEAN global      /* sort posterior probabilities over all sequences? */
		   )
{
  long maxSeqLength;        /* maximum sequence length */
  int seqNum;               /* number of current sequence */
  int numSeqs;              /* number of sequences */
  int siteNum;              /* number of current site */
  int maxNumSites;          /* maximum number of sites */
  POSTPROB *postProbs;      /* posterior probabilities to be sorted */
  POSTPROB *maxPostProbs;   /* sorted posterior probabilities */

  numSeqs       = dataSet->numSeqs;
  maxSeqLength  = dataSet->maxSeqLength;
  maxNumSites   = dataSet->maxNumSites;
  postProbs     = dataSet->postProbs;
  maxPostProbs  = dataSet->maxPostProbs;

  if (global){
    qsort((void*) postProbs,maxSeqLength*numSeqs,sizeof(POSTPROB),compPostProbs);
    for (siteNum=0; siteNum<maxNumSites; siteNum++){
      maxPostProbs[siteNum].seqNum     = postProbs[siteNum].seqNum;
      maxPostProbs[siteNum].start      = postProbs[siteNum].start;
      maxPostProbs[siteNum].prob       = postProbs[siteNum].prob;
      maxPostProbs[siteNum].revComp    = postProbs[siteNum].revComp;
    }
  }else{
    for (seqNum=0; seqNum<numSeqs; seqNum++){
      qsort((void*) (postProbs + seqNum*maxSeqLength),maxSeqLength,sizeof(POSTPROB),compPostProbs);
      maxPostProbs[seqNum].seqNum     = postProbs[PP(seqNum,0)].seqNum;
      maxPostProbs[seqNum].start      = postProbs[PP(seqNum,0)].start;
      maxPostProbs[seqNum].prob       = postProbs[PP(seqNum,0)].prob;
      maxPostProbs[seqNum].revComp    = postProbs[PP(seqNum,0)].revComp;
    }
    qsort((void*) maxPostProbs, numSeqs,sizeof(POSTPROB),compPostProbs);
  }

  return(1);
}
