/*****************************************************************************
*			     					             *
*      cosmo - Constrained search for motifs in DNA sequences                *
*      Copyright (C) 2006  Oliver Bembom (bembom@berkeley.edu)               *
*                                                                            *
*      This file is part of cosmo.                                           *
*                                                                            *
*      cosmo is free software; you can redistribute it and/or                *
*      modify it under the terms of the GNU Lesser General Public            *
*      License as published by the Free Software Foundation; either          *
*      version 2.1 of the License, or (at your option) any later version.    *
*                                                                            *
*      cosmo is distributed in the hope that it will be useful,              *
*      but WITHOUT ANY WARRANTY; without even the implied warranty of        *
*      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU     *
*      Lesser General Public License for more details.                       *
*                                                                            *
*      You should have received a copy of the GNU Lesser General Public      *
*      License along with this program; if not, write to the Free Software   *
*      Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,            *
*      MA  02110-1301, USA.                                                  *
*                                                                            *
*                                                                            *
*****************************************************************************/


/*****************************************************************************
*                                                                            *
*      This module contains various input/output functions for cosmo for     *
*      interfacing with R.                                                   *
*                                                                            *
*****************************************************************************/

#include <R_ext/PrtUtil.h>
#include <R_ext/Memory.h>
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <float.h>
#include "macros.h"
#include "cosmo.h"
#include "alloc.h"
#include "rio.h"
#include "matLib.h"
#include "eval.h"
#include "io.h"
#include "posterior.h"

  
/**********************************************************************/
/*
  readSeqsFromVec

  Read sequences in from matrix supplied as input from R
  Sequences are passed in stacked in one long vector, stored in the 
  format A=1, C=2, G=3, T=4, with each sequence allotted the space
  of the longest sequence and additional entries filled with 0s.
  

*/
/**********************************************************************/
int readSeqsFromVec(
		    int *seqs,          /* vector that stores sequences */
		    DATASET *dataSet    /* the dataset */
		    )
{
  int seqNum;        /* number of current sequence */
  int numSeqs;       /* number of sequences */
  long pos;          /* position in sequence */
  long seqLength;    /* sequence length */
  long maxSeqLength; /* length of longest sequence */
  int *seq;          /* pointer to sequence */
  int minWidth;      /* minimum motif width to consider */
  int maxWidth;      /* maximum motif width to consider */
  int maxFold;       /* maximum fold to consider */
  SAMPLE *sample;    /* current sample */

  minWidth                = dataSet->minWidth;
  maxWidth                = dataSet->maxWidth;
  numSeqs                 = dataSet->numSeqs;
  maxSeqLength            = dataSet->maxSeqLength;

  /* allocate memory for samples */
  dataSet->samples = sampleMalloc(dataSet->numSeqs);

  /* get sequence lengths: sequence ends at first 0 entry in vector */
  for (seqNum=0; seqNum<numSeqs; seqNum++){
    seq       = seqs + seqNum*maxSeqLength;
    pos       = 0;

    while (pos < maxSeqLength && seq[pos]!=0) pos++;

    dataSet->samples[seqNum].length  = pos;
  }


  /* set minNumSites and maxNumSites if necessary */
  if (!dataSet->minNumSites) dataSet->minNumSites = 2;
  if (!dataSet->maxNumSites) dataSet->maxNumSites = MIN(50, 5*dataSet->numSeqs);

  /* set Folds */
  dataSet->backFold     = MIN(dataSet->backFold, dataSet->numSeqs);
  dataSet->intCrit.fold = MIN(dataSet->intCrit.fold, dataSet->numSeqs);
  dataSet->wCrit.fold   = MIN(dataSet->wCrit.fold, dataSet->numSeqs);
  dataSet->modCrit.fold = MIN(dataSet->modCrit.fold, dataSet->numSeqs);
  dataSet->conCrit.fold = MIN(dataSet->conCrit.fold, dataSet->numSeqs);
  if (dataSet->conCrit.crit==pwmCV)   dataSet->conCrit.fold = MIN(dataSet->conCrit.fold, 2);

  /* allocate memorty for excludeCv vectors */
  maxFold               = MAX(MAX(MAX(dataSet->intCrit.fold, dataSet->wCrit.fold),dataSet->modCrit.fold),dataSet->conCrit.fold);
  dataSet->excludeCv    = IntMalloc(maxFold);
  dataSet->excludeCv2   = IntMalloc(maxFold);

  /* allocate memory for cumProb */
  dataSet->cumProb   = DoubleMalloc(dataSet->maxSeqLength+1);
  
  /* allocate memory for postProbs and maxPostProbs */
  dataSet->postProbs    = postProbMalloc(dataSet->numSeqs*dataSet->maxSeqLength);
  dataSet->maxPostProbs = postProbMalloc(dataSet->maxNumSites);

  /* allocate memory for intMotProbs */
  dataSet->intMotProbs1 = intMotProbMalloc((dataSet->numSeqs)*(dataSet->maxSeqLength)*(1 + dataSet->revComp));
  dataSet->intMotProbs2 = intMotProbMalloc((dataSet->numSeqs)*(dataSet->maxSeqLength)*(1 + dataSet->revComp));
  dataSet->maxMotProbs  = intMotProbMalloc((dataSet->numSeqs)*(dataSet->maxSeqLength)*(1 + dataSet->revComp));

  /* allocate memory for logLiks */
  dataSet->logLiks      = DoubleMalloc(dataSet->numSeqs);


  /* allocate memory for various arrays in samples */
  for (seqNum=0; seqNum<dataSet->numSeqs; seqNum++){
    seqLength           = dataSet->samples[seqNum].length;
    dataSet->totalNucs += seqLength;

    if (DEBUG>=3){
      Rprintf("initSamples: seq %d seqLength = %ld\n", seqNum, seqLength);
    }


    dataSet->samples[seqNum].numCuts        = IntMalloc(maxWidth-minWidth+1);
    dataSet->samples[seqNum].seq            = IntMalloc(seqLength);
    dataSet->samples[seqNum].rcSeq          = IntMalloc(seqLength);

    dataSet->samples[seqNum].logCumBack     = DoubleMalloc(seqLength+1);
    dataSet->samples[seqNum].backLR         = DoubleMalloc(seqLength*(maxWidth-minWidth+1));
    dataSet->samples[seqNum].cumProbBack    = DoubleMalloc(seqLength*(maxWidth-minWidth+1));

    dataSet->samples[seqNum].mProbs1        = DoubleMalloc(seqLength);
    dataSet->samples[seqNum].rcmProbs1      = DoubleMalloc(seqLength);
    dataSet->samples[seqNum].mProbs2        = DoubleMalloc(seqLength);
    dataSet->samples[seqNum].rcmProbs2      = DoubleMalloc(seqLength);

    dataSet->samples[seqNum].seqProbs       = DoubleMalloc(seqLength);
    dataSet->samples[seqNum].rcSeqProbs     = DoubleMalloc(seqLength);

  }



  /* read in sequences */
  for (seqNum=0; seqNum<dataSet->numSeqs; seqNum++){
    sample     = dataSet->samples + seqNum;
    seqLength  = sample->length;
    seq        = sample->seq;

    /* Sequences stored internally as 0-4 but read in as 1-5 */
    /* 4/5 is the unknown character 'X' */
    for (pos=0; pos<seqLength;pos++) seq[pos] = seqs[seqNum*maxSeqLength + pos] - 1;
  }

  /* get reverse complement if necessary */
  if (dataSet->revComp){
    for (seqNum=0;seqNum<dataSet->numSeqs;seqNum++){
      sample = &(dataSet->samples[seqNum]);

      for (pos=0;pos<sample->length;pos++){
	if (sample->seq[sample->length - pos - 1]<4){
	  sample->rcSeq[pos] = 3 - sample->seq[sample->length - pos - 1];
	}else{
	  /* unknown character 'X' */
	  sample->rcSeq[pos] = 4;
	}
      }
    }
  }


  /* get background frequencies */
  if (!getBackFreqs(dataSet)) return(0);


  /* Initialize the probability tables for the objective function.  */
  /* Get for 2 and up because weighting may cause < min_nsites sites */
  init_llr_pv_tables(2, dataSet->maxNumSites, dataSet->nucFreq);

 
  if (DEBUG>=3){
    Rprintf("readSeqsFromVec:\n");
    Rprintf("numSeqs: %ld\n",dataSet->numSeqs);  
	
    for (seqNum=0;seqNum<dataSet->numSeqs;seqNum++){
      sample = &(dataSet->samples[seqNum]);

      Rprintf("Sequence %ld: ", seqNum);
      Rprintf("Length: %d\n", sample->length);
      PrintIntMatrix(sample->seq,1,sample->length);

      if (dataSet->revComp){
	Rprintf("reverse complement: ");
	PrintIntMatrix(sample->rcSeq,1,sample->length);
      }
    }
  }

  return(1);
}


/**********************************************************************/
/*
  readBackSeqsFromVec

  Read background sequences in from matrix supplied as input from R

  Sequences are passed in stacked in one long vector, stored in the 
  format A=1, C=2, G=3, T=4, with each sequence allotted the space
  of the longest sequence and additional entries filled with 0s.

*/
/**********************************************************************/
int readBackSeqsFromVec(
			int *seqs,           /* vector of sequences */
			DATASET *dataSet     /* the dataset */
			)
{
  int seqNum;           /* number of current sequence */
  int numSeqs;          /* number of background sequences */
  long pos;             /* position in current sequence */ 
  long seqLength;       /* length of sequence */
  long maxSeqLength;    /* length of longest background sequence */
  int *seq;             /* pointer to sequence */
  int minWidth;         /* minimum motif width to consider */
  int maxWidth;         /* maximum motif width to consider */
  SAMPLE *sample;       /* current sample */

  minWidth                = dataSet->minWidth;
  maxWidth                = dataSet->maxWidth;
  numSeqs                 = dataSet->numBackSeqs;
  maxSeqLength            = dataSet->maxBackSeqLength;

  /* allocate memory for samples */
  dataSet->backSamples = sampleMalloc(dataSet->numBackSeqs);

  /* get sequence lengths: sequence ends at first 0 entry in vector */
  for (seqNum=0; seqNum<numSeqs; seqNum++){
    seq       = seqs + seqNum*maxSeqLength;
    pos       = 0;

    while (pos < maxSeqLength && seq[pos]!=0) pos++;

    dataSet->backSamples[seqNum].length  = pos;

    /* allocate memory for sequence */
    dataSet->backSamples[seqNum].seq            = IntMalloc(pos);
  }



  /* read in sequences */
  for (seqNum=0; seqNum<dataSet->numBackSeqs; seqNum++){
    sample     = dataSet->backSamples + seqNum;
    seqLength  = sample->length;
    seq        = sample->seq;

    /* Sequences stored internally as 0-4 but read in as 1-5 */
    for (pos=0; pos<seqLength;pos++) seq[pos] = seqs[seqNum*maxSeqLength + pos] - 1;
    
  }      

  return(1);
}

/**********************************************************************/
/*
  readTransMatFromVec

  Reads in transition matrix from vector as input from R.
  Transition matrices are stored stacked in one vector, in decreasing
  order of the order. 


*/
/**********************************************************************/
int readTransMatFromVec(
			double *transMats, /* transition matrices */
			DATASET *dataSet   /* the dataset */
		       )
{
  int row;              /* row in transition matrix */
  int col;              /* column in transition matrix */
  int order;            /* current order */
  int subOrder;         /* lower order */
  long numRows;         /* number of rows altogether */
  long index;           /* index into matrix */
  double *transMat;     /* transition matrix */
  long addRows;         /* number of rows to add for previous orders */


  order = dataSet->order;

  dataSet->transMats = p2DoubleMalloc(order+1);
  for (subOrder=0; subOrder<=order; subOrder++) dataSet->transMats[subOrder] = DoubleMalloc(pow(4,subOrder+1));

  if (DEBUG>=2) Rprintf("readTransMatFromVec: reading transition matrix for background Markov model with order = %d\n",  order);
  addRows = 0;
  for (subOrder=order; subOrder>=0; subOrder--){

    transMat = dataSet->transMats[subOrder];
    numRows  = pow(4,subOrder);

    for (row=0; row<numRows; row++){
      for (col=0; col<4; col++){
	index = 4*row + col;
	transMat[index] = transMats[addRows + index];

	/* transform to logs */
	transMat[index] = LOG(transMat[index]);
      }
    }


    Rprintf("transMat for order = %d\n", subOrder);
    PrintDoubleMatrix2File(stderr,transMat,4,numRows);
    Rprintf("\n");
  }

  return(1);
}



/**********************************************************************/
/*
  output results to various vectors to be passed to R

*/
/**********************************************************************/
int output2vec(
	       DATASET *dataSet,      /* the dataset */
	       int *selOrder,         /* chosen order of background model */
	       double *klDivs,        /* Kullback-Leibler divergences for candidate orders of background model */
	       double *estTransMat,   /* estimated transition matrices */
	       double *wCrits,        /* width criteria of candidate models */
	       double *modCrits,      /* model criteria of candidate models */
	       double *conCrits,      /* constraint set criteria of candidate models */
	       int *selWidth,         /* chosen motif width */
 	       int *selModel,         /* chosen model type */
	       int *selConSet,        /* chosen constraint set number */
	       int *selNumSites,      /* chosen number of motif occurrences */
	       int *numModels,        /* number of candidate models */
	       double *pwm,           /* estimated position weight matrix */
	       double *motProbs,      /* posterior probabilities of motif occurrences */
	       int *motRevComp,       /* is posterior probability greater on reverse strand? */
	       int *alignSeqs,        /* sequences in which the motifs can be found */
	       int *alignStarts,      /* starting positions of the the motifs */
	       double *eVal           /* the evalue of the alignment */
	       )
{
  THETA *theta;            /* parameter estimates */
  int pos;                 /* position */
  int nuc;                 /* nucleotide */
  int seqNum;              /* number of sequence */
  int index;               /* index */
  int i;                   /* index */
  int subOrder;            /* a particular order */
  POSTPROB *postProbs;     /* posterior probabilities */
  POSTPROB *maxPostProbs;  /* sorted maximal posterior probabilities */
  double *transMat;        /* transition matrix for given order */
  int addRows;             /* number of rows to add for previous transition matrices */
  int numRows;             /* number of rows in this transition matrix */
  int row;                 /* current row */
  int col;                 /* current column */
  int width;               /* motif width */
  long maxSeqLength;       /* maximum sequence length */
  BOOLEAN global;          /* do posterior probabilities need to be sorted globally or for each sequence separately? */

  postProbs     = dataSet->postProbs;
  maxPostProbs  = dataSet->maxPostProbs;
  theta         = dataSet->theta;
  maxSeqLength  = dataSet->maxSeqLength;
  global        = theta->mType==TCM || theta->mType==OVER || theta->mType==CUT;

  *selOrder     = dataSet->order;
  *selWidth     = theta->width;
  width         = theta->width;
  *selModel     = dataSet->crits.mType + 1;
  *selConSet    = dataSet->crits.conSetNum + 1;

  *selNumSites  = theta->numSites;
  *numModels    = dataSet->numModels;
  *eVal         = exp(theta->logEval);

  /* background Kullback-Leibler divergences */
  for (i=0; i<dataSet->maxOrder; i++) klDivs[i] = dataSet->backKL[i];

  /* model selection criteria */
  for (i=0; i<dataSet->numModels; i++){

    if (dataSet->wCrit.crit==eval){
      wCrits[i]   = dataSet->allCrits[i].wInit   ? exp(dataSet->allCrits[i].wCrit)     : 0;
    }else if (dataSet->wCrit.crit==lik){
      wCrits[i]   = dataSet->allCrits[i].wInit   ? -dataSet->allCrits[i].wCrit     : 0;
    }else{
      wCrits[i]   = dataSet->allCrits[i].wInit   ? dataSet->allCrits[i].wCrit     : 0;
    }
 
    if (dataSet->modCrit.crit==eval){
      modCrits[i]   = dataSet->allCrits[i].modInit   ? exp(dataSet->allCrits[i].modCrit)     : 0;
    }else if (dataSet->modCrit.crit==lik){
      modCrits[i]   = dataSet->allCrits[i].modInit   ? -dataSet->allCrits[i].modCrit     : 0;
    }else{
      modCrits[i]   = dataSet->allCrits[i].modInit   ? dataSet->allCrits[i].modCrit     : 0;
    }

    if (dataSet->conCrit.crit==eval){
      conCrits[i]   = dataSet->allCrits[i].conInit   ? exp(dataSet->allCrits[i].conCrit)     : 0;
    }else if (dataSet->conCrit.crit==lik){
      conCrits[i]   = dataSet->allCrits[i].conInit   ? -dataSet->allCrits[i].conCrit     : 0;
    }else{
      conCrits[i]   = dataSet->allCrits[i].conInit   ? dataSet->allCrits[i].conCrit     : 0;
    }
  }


 
  /* print position weight matrix */
  for(nuc=0;nuc<4;nuc++){
    for(pos=0;pos<(*selWidth);pos++){
      /* Rprintf("Entering pos %d nuc %d of PWM\n", pos, nuc); */
      pwm[PWM(pos,nuc)] = theta->pwm[PWM(pos,nuc)];
    }
  }

  /* estimated transition matrices */
  addRows = 0;
  for (subOrder=dataSet->order; subOrder>=0; subOrder--){

    transMat = dataSet->transMats[subOrder];
    numRows  = pow(4,subOrder);

    for (row=0; row<numRows; row++){
      for (col=0; col<4; col++){
	estTransMat[4*(addRows+row) + col] = exp(transMat[4*row+col]);
      }
    }
    addRows += numRows;
  }

  
  /* get posterior probabilities */
  if (!getPostProbs(theta, dataSet)) return(0);

  index = 0;
  for (seqNum=0; seqNum<dataSet->numSeqs; seqNum++){
    for (pos=0; pos<maxSeqLength;pos++){
      motProbs[index] = postProbs[index].prob;
      motRevComp[index] = postProbs[index].revComp;
      index++;
    }
  }

  /* sort posterior probabilities */
  if (!sortPostProbs(dataSet, global)) return(0);
  for (index=0; index<*selNumSites;index++){
    alignStarts[index]  = maxPostProbs[index].start + 1;
    alignSeqs[index]    = maxPostProbs[index].seqNum + 1;
  }

  return(1);
}


